#!/usr/bin/env python3

r'''Tests projections, unprojections, distortions, etc'''

import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils


m               = mrcal.cameramodel(f"{testdir}/data/cam0.opencv8.cameramodel")
W,H             = m.imagersize()
intrinsics_core = m.intrinsics()[1][:4]

testutils.confirm_equal( mrcal.scale_focal__best_pinhole_fit(m, None),
                         1.0,
                         msg = 'scale_focal__best_pinhole_fit')


def fit_check(scale_focal, intrinsics, v,
              W                        = W,
              H                        = H,
              scale_imagersize_pinhole = 1.0,
              eps                      = 1e-2):
    r'''Makes sure projected vectors fit into the imager perfectly

    I'm given a number of points in the camera coords. These much project such
    that

    - All projected points lie INSIDE the imager
    - At least one point lies exactly on the imager boundary

    '''
    intrinsics = intrinsics.copy()
    intrinsics[:2] *= scale_focal
    intrinsics     *= scale_imagersize_pinhole

    q = mrcal.project(v, 'LENSMODEL_PINHOLE', intrinsics)

    if any( q[:,0] < -eps )    or \
       any( q[:,0] > W-1+eps ) or \
       any( q[:,1] < -eps )    or \
       any( q[:,1] > H-1+eps ):
        return ": Some points lie out of bounds"

    on_left_edge   = np.abs(q[:,0]        ) < eps
    on_right_edge  = np.abs(q[:,0] - (W-1)) < eps
    on_top_edge    = np.abs(q[:,1]        ) < eps
    on_bottom_edge = np.abs(q[:,1] - (H-1)) < eps

    if not ( \
             any(on_left_edge)  or \
             any(on_top_edge)   or \
             any(on_right_edge) or \
             any(on_bottom_edge) ):
        return ": No points lie on the edge"

    # all good
    return ''


err_msg = \
    fit_check( mrcal.scale_focal__best_pinhole_fit(m, 'corners'),
               intrinsics_core,
               mrcal.unproject( np.array(((  0 ,  0),
                                          (W-1,   0),
                                          (  0, H-1),
                                          (W-1, H-1)), dtype=float),
                                *m.intrinsics()),)
testutils.confirm( err_msg == '',
                   msg = 'scale_focal__best_pinhole_fit' + err_msg)

err_msg = \
    fit_check( mrcal.scale_focal__best_pinhole_fit(m, 'centers-horizontal'),
               intrinsics_core,
               mrcal.unproject( np.array(((  0, (H-1.)/2.),
                                          (W-1, (H-1.)/2.)), dtype=float),
                                *m.intrinsics()),)
testutils.confirm( err_msg == '',
                   msg = 'scale_focal__best_pinhole_fit' + err_msg)

err_msg = \
    fit_check( mrcal.scale_focal__best_pinhole_fit(m, 'centers-vertical'),
               intrinsics_core,
               mrcal.unproject( np.array((((W-1.)/2.,   0.),
                                          ((W-1.)/2., H-1.)), dtype=float),
                                *m.intrinsics()),)
testutils.confirm( err_msg == '',
                   msg = 'scale_focal__best_pinhole_fit' + err_msg)

err_msg = \
    fit_check( mrcal.scale_focal__best_pinhole_fit(m,
                                                     np.array((((W-1.)/2.,   0.),
                                                               ((W-1.)/2., H-1.)), dtype=float)),
               intrinsics_core,
               mrcal.unproject( np.array((((W-1.)/2.,   0.),
                                          ((W-1.)/2., H-1.)), dtype=float),
                                *m.intrinsics()),)
testutils.confirm( err_msg == '',
                   msg = 'scale_focal__best_pinhole_fit' + err_msg)

testutils.finish()



#!/usr/bin/python3

# Copyright (c) 2017-2023 California Institute of Technology ("Caltech"). U.S.
# Government sponsorship acknowledged. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

'''Routines for transformation of images

All functions are exported into the mrcal module. So you can call these via
mrcal.image_transforms.fff() or mrcal.fff(). The latter is preferred.

'''

import numpy as np
import numpysane as nps
import sys
import re
import mrcal

def scale_focal__best_pinhole_fit(model, fit):
    r'''Compute the optimal focal-length scale for reprojection to a pinhole lens

SYNOPSIS

    model = mrcal.cameramodel('from.cameramodel')

    lensmodel,intrinsics_data = model.intrinsics()

    scale_focal = mrcal.scale_focal__best_pinhole_fit(model,
                                                      'centers-horizontal')

    intrinsics_data[:2] *= scale_focal

    model_pinhole = \
        mrcal.cameramodel(intrinsics = ('LENSMODEL_PINHOLE',
                                        intrinsics_data[:4]),
                          imagersize = model.imagersize(),
                          rt_cam_ref = model.rt_cam_ref() )

Many algorithms work with images assumed to have been captured with a pinhole
camera, even though real-world lenses never fit a pinhole model. mrcal provides
several functions to remap images captured with non-pinhole lenses into images
of the same scene as if they were observed by a pinhole lens. When doing this,
we're free to choose all of the parameters of this pinhole lens model, and this
function allows us to pick the best scaling on the focal-length parameter.

The focal length parameters serve as a "zoom" factor: changing these parameters
can increase the resolution of the center of the image, at the expense of
cutting off the edges. This function computes the best focal-length scaling, for
several possible meanings of "best".

I assume an output pinhole model that has pinhole parameters

    (k*fx, k*fy, cx, cy)

where (fx, fy, cx, cy) are the parameters from the input model, and k is the
scaling we compute.

This function looks at some points on the edge of the input image. I choose k so
that all of these points end up inside the pinhole-reprojected image, leaving
the worst one at the edge. The set of points I look at are specified in the
"fit" argument.

ARGUMENTS

- model: a mrcal.cameramodel object for the input lens model

- fit: which pixel coordinates in the input image must project into the output
  pinhole-projected image. The 'fit' argument must be one of

  - a numpy array of shape (N,2) where each row is a pixel coordinate in the
    input image

  - "corners": each of the 4 corners of the input image must project into the
    output image

  - "centers-horizontal": the two points at the left and right edges of the
    input image, half-way vertically, must both project into the output image

  - "centers-vertical": the two points at the top and bottom edges of the input
    image, half-way horizontally, must both project into the output image

RETURNED VALUE

A scalar scale_focal that can be passed to pinhole_model_for_reprojection()

    '''

    if fit is None: return 1.0

    WH  = np.array(model.imagersize(), dtype=float)
    W,H = WH

    if type(fit) is np.ndarray:
        q_edges = fit
    elif type(fit) is str:
        if fit == 'corners':
            q_edges = np.array(((  0.,   0.),
                                (  0., H-1.),
                                (W-1., H-1.),
                                (W-1.,   0.)))
        elif fit == 'centers-horizontal':
            q_edges = np.array(((0,    (H-1.)/2.),
                                (W-1., (H-1.)/2.)))
        elif fit == 'centers-vertical':
            q_edges = np.array((((W-1.)/2., 0,   ),
                                ((W-1.)/2., H-1.,)))
        else:
            raise Exception("fit must be either None or a numpy array or one of ('corners','centers-horizontal','centers-vertical')")
    else:
        raise Exception("fit must be either None or a numpy array or one of ('corners','centers-horizontal','centers-vertical')")

    lensmodel,intrinsics_data = model.intrinsics()

    v_edges = mrcal.unproject(q_edges, lensmodel, intrinsics_data)

    if not mrcal.lensmodel_metadata_and_config(lensmodel)['has_core']:
        raise Exception("This currently works only with models that have an fxfycxcy core")
    fxy = intrinsics_data[ :2]
    cxy = intrinsics_data[2:4]

    # I have points in space now. My scaled pinhole camera would map these to
    # (k*fx*x/z+cx, k*fy*y/z+cy). I pick a k to make sure that this is in-bounds
    # for all my points, and that the points occupy as large a chunk of the
    # imager as possible. I can look at just the normalized x and y. Just one of
    # the query points should land on the edge; the rest should be in-bounds

    normxy_edges = v_edges[:,:2] / v_edges[:,(2,)]
    normxy_min   = (      - cxy) / fxy
    normxy_max   = (WH-1. - cxy) / fxy

    # Each query point will imply a scale to just fit into the imager I take the
    # most conservative of these. For each point I look at the normalization sign to
    # decide if I should be looking at the min or max edge. And for each I pick the
    # more conservative scale

    # start out at an unreasonably high scale. The data will cut this down
    scale = 1e6
    for p in normxy_edges:
        for ixy in range(2):
            if p[ixy] > 0: scale = np.min((scale,normxy_max[ixy]/p[ixy]))
            else:          scale = np.min((scale,normxy_min[ixy]/p[ixy]))
    return scale


def pinhole_model_for_reprojection(model_from,
                                   fit         = None,
                                   *,
                                   scale_focal = None,
                                   scale_image = None):

    r'''Generate a pinhole model suitable for reprojecting an image

SYNOPSIS

    model_orig = mrcal.cameramodel("xxx.cameramodel")
    image_orig = mrcal.load_image("image.jpg")

    model_pinhole = mrcal.pinhole_model_for_reprojection(model_orig,
                                                         fit = "corners")

    mapxy = mrcal.image_transformation_map(model_orig, model_pinhole,
                                           intrinsics_only = True)

    image_undistorted = mrcal.transform_image(image_orig, mapxy)

Many algorithms work with images assumed to have been captured with a pinhole
camera, even though real-world lenses never fit a pinhole model. mrcal provides
several functions to remap images captured with non-pinhole lenses into images
of the same scene as if they were observed by a pinhole lens. When doing this,
we're free to choose all of the parameters of this pinhole lens model. THIS
function produces the pinhole camera model based on some guidance in the
arguments, and this model can then be used to "undistort" images.

ARGUMENTS

- model_from: the mrcal.cameramodel object used to build the pinhole model. We
  use the intrinsics as the baseline, and we copy the extrinsics to the
  resulting pinhole model.

- fit: optional specification for focal-length scaling. By default we use the
  focal length values from the input model. This is either a numpy array of
  shape (...,2) containing pixel coordinates that the resulting pinhole model
  must represent, or one of ("corners","centers-horizontal","centers-vertical").
  See the docstring for scale_focal__best_pinhole_fit() for details. Exclusive
  with 'scale_focal'

- scale_focal: optional specification for focal-length scaling. By default we
  use the focal length values from the input model. If given, we scale the input
  focal lengths by the given value. Exclusive with 'fit'

- scale_image: optional specification for the scaling of the image size. By
  default the output model represents an image of the same resolution as the
  input model. If we want something else, the scaling can be given here.

RETURNED VALUE

A mrcal.cameramodel object with lensmodel = LENSMODEL_PINHOLE corresponding to
the input model.

    '''

    if scale_focal is None:

        if fit is not None:
            if isinstance(fit, np.ndarray):
                if fit.shape[-1] != 2:
                    raise Exception("'fit' is an array, so it must have shape (...,2)")
                fit = nps.atleast_dims(fit, -2)
                fit = nps.clump(fit, n=len(fit.shape)-1)
                # fit now has shape (N,2)

            elif re.match("^(corners|centers-horizontal|centers-vertical)$", fit):
                # this is valid. nothing to do
                pass
            else:
                raise Exception("'fit' must be an array of shape (...,2) or one of ('corners','centers-horizontal','centers-vertical')",
                      file=sys.stderr)
                sys.exit(1)

        scale_focal = mrcal.scale_focal__best_pinhole_fit(model_from, fit)

    else:
        if fit is not None:
            raise Exception("At most one of 'scale_focal' and 'fit' may be non-None")

    # I have some scale_focal now. I apply it
    lensmodel,intrinsics_data = model_from.intrinsics()
    imagersize                = model_from.imagersize()

    if not mrcal.lensmodel_metadata_and_config(lensmodel)['has_core']:
        raise Exception("This currently works only with models that have an fxfycxcy core")
    cx,cy = intrinsics_data[2:4]
    intrinsics_data[:2] *= scale_focal

    if scale_image is not None:
        # Now I apply the imagersize scale. The center of the imager should
        # unproject to the same point:
        #
        #   (q0 - cxy0)/fxy0 = v = (q1 - cxy1)/fxy1
        #   ((WH-1)/2 - cxy) / fxy = (((ki*WH)-1)/2 - kc*cxy) / (kf*fxy)
        #
        # The focal lengths scale directly: kf = ki
        #   ((WH-1)/2 - cxy) / fxy = (((ki*WH)-1)/2 - kc*cxy) / (ki*fxy)
        #   (WH-1)/2 - cxy = (((ki*WH)-1)/2 - kc*cxy) / ki
        #   (WH-1)/2 - cxy = (WH-1/ki)/2 - kc/ki*cxy
        #   -1/2 - cxy = (-1/ki)/2 - kc/ki*cxy
        #   1/2 + cxy = 1/(2ki) + kc/ki*cxy
        # -> kc = (1/2 + cxy - 1/(2ki)) * ki / cxy
        #       = (ki + 2*cxy*ki - 1) / (2 cxy)
        #
        # Sanity check: cxy >> 1: ki+2*cxy*ki = ki*(1+2cxy) ~ 2*cxy*ki
        #                         2*cxy*ki - 1 ~ 2*cxy*ki
        #               -> kc ~ 2*cxy*ki /( 2 cxy ) = ki. Yes.
        # Looks like I scale cx and cy separately.
        imagersize[0] = round(imagersize[0]*scale_image)
        imagersize[1] = round(imagersize[1]*scale_image)
        kfxy = scale_image
        kcx  = (kfxy + 2.*cx*kfxy - 1.) / (2. * cx)
        kcy  = (kfxy + 2.*cy*kfxy - 1.) / (2. * cy)

        intrinsics_data[:2] *= kfxy
        intrinsics_data[2]  *= kcx
        intrinsics_data[3]  *= kcy

    return \
        mrcal.cameramodel( intrinsics            = ('LENSMODEL_PINHOLE',intrinsics_data[:4]),
                           rt_cam_ref = model_from.rt_cam_ref(),
                           imagersize            = imagersize )


def image_transformation_map(model_from, model_to,
                             *,
                             intrinsics_only                   = False,
                             distance                          = None,
                             plane_n                           = None,
                             plane_d                           = None,
                             mask_valid_intrinsics_region_from = False):

    r'''Compute a reprojection map between two models

SYNOPSIS

    model_orig = mrcal.cameramodel("xxx.cameramodel")
    image_orig = mrcal.load_image("image.jpg")

    model_pinhole = mrcal.pinhole_model_for_reprojection(model_orig,
                                                         fit = "corners")

    mapxy = mrcal.image_transformation_map(model_orig, model_pinhole,
                                           intrinsics_only = True)

    image_undistorted = mrcal.transform_image(image_orig, mapxy)

    # image_undistorted is now a pinhole-reprojected version of image_orig

Returns the transformation that describes a mapping

- from pixel coordinates of an image of a scene observed by model_to
- to pixel coordinates of an image of the same scene observed by model_from

This transformation can then be applied to a whole image by calling
mrcal.transform_image().

This function returns a transformation map in an (Nheight,Nwidth,2) array. The
image made by model_to will have shape (Nheight,Nwidth). Each pixel (x,y) in
this image corresponds to a pixel mapxy[y,x,:] in the image made by model_from.

One application of this function is to validate the models in a stereo pair. For
instance, reprojecting one camera's image at distance=infinity should produce
exactly the same image that is observed by the other camera when looking at very
far objects, IF the intrinsics and rotation are correct. If the images don't
line up well, we know that some part of the models is off. Similarly, we can use
big planes (such as observations of the ground) and plane_n, plane_d to validate.

This function has several modes of operation:

- intrinsics, extrinsics

  Used if not intrinsics_only and \
          plane_n is None     and \
          plane_d is None

  This is the default. For each pixel in the output, we use the full model to
  unproject a given distance out, and then use the full model to project back
  into the other camera.

- intrinsics only

  Used if intrinsics_only and \
          plane_n is None and \
          plane_d is None

  Similar, but the extrinsics are ignored. We unproject the pixels in one model,
  and project the into the other camera. The two camera coordinate systems are
  assumed to line up perfectly

- plane

  Used if plane_n is not None and
          plane_d is not None

  We map observations of a given plane in camera FROM coordinates coordinates to
  where this plane would be observed by camera TO. We unproject each pixel in
  one camera, compute the intersection point with the plane, and project that
  intersection point back to the other camera. This uses ALL the intrinsics,
  extrinsics and the plane representation. The plane is represented by a normal
  vector plane_n, and the distance to the normal plane_d. The plane is all
  points p such that inner(p,plane_n) = plane_d. plane_n does not need to be
  normalized; any scaling is compensated in plane_d.

ARGUMENTS

- model_from: the mrcal.cameramodel object describing the camera used to capture
  the input image. We always use the intrinsics. if not intrinsics_only: we use
  the extrinsics also

- model_to: the mrcal.cameramodel object describing the camera that would have
  captured the image we're producing. We always use the intrinsics. if not
  intrinsics_only: we use the extrinsics also

- intrinsics_only: optional boolean, defaulting to False. If False: we respect
  the relative transformation in the extrinsics of the camera models.

- distance: optional value, defaulting to None. Used only if not
  intrinsics_only. We reproject points in space a given distance out. If
  distance is None (the default), we look out to infinity. This is equivalent to
  using only the rotation component of the extrinsics, ignoring the translation.

- plane_n: optional numpy array of shape (3,); None by default. If given, we
  produce a transformation to map observations of a given plane to the same
  pixels in the source and target images. This argument describes the normal
  vector in the coordinate system of model_from. The plane is all points p such
  that inner(p,plane_n) = plane_d. plane_n does not need to be normalized; any
  scaling is compensated in plane_d. If given, plane_d should be given also, and
  intrinsics_only should be False. if given, we use the full intrinsics and
  extrinsics of both camera models

- plane_d: optional floating-point value; None by default. If given, we produce
  a transformation to map observations of a given plane to the same pixels in
  the source and target images. The plane is all points p such that
  inner(p,plane_n) = plane_d. plane_n does not need to be normalized; any
  scaling is compensated in plane_d. If given, plane_n should be given also, and
  intrinsics_only should be False. if given, we use the full intrinsics and
  extrinsics of both camera models

- mask_valid_intrinsics_region_from: optional boolean defaulting to False. If
  True, points outside the valid-intrinsics region in the FROM image are set to
  black, and thus do not appear in the output image

RETURNED VALUE

A numpy array of shape (Nheight,Nwidth,2) where Nheight and Nwidth represent the
imager dimensions of model_to. This array contains 32-bit floats, as required by
cv2.remap() (the function providing the internals of mrcal.transform_image()).
This array can be passed to mrcal.transform_image()

    '''

    if (plane_n is      None and plane_d is not None) or \
       (plane_n is not  None and plane_d is     None):
        raise Exception("plane_n and plane_d should both be None or neither should be None")
    if plane_n is not None and \
       intrinsics_only:
        raise Exception("We're looking at remapping a plane (plane_d, plane_n are not None), so intrinsics_only should be False")

    if distance is not None and \
       (plane_n is not None or intrinsics_only):
        raise Exception("'distance' makes sense only without plane_n/plane_d and without intrinsics_only")

    if intrinsics_only:
        Rt_to_from = None
    else:
        Rt_to_from = mrcal.compose_Rt(model_to.  Rt_cam_ref(),
                                      model_from.Rt_ref_cam())

    lensmodel_from,intrinsics_data_from = model_from.intrinsics()
    lensmodel_to,  intrinsics_data_to   = model_to.  intrinsics()

    if re.match("LENSMODEL_OPENCV",lensmodel_from) and \
       lensmodel_to == "LENSMODEL_PINHOLE"         and \
       plane_n is None                             and \
       not mask_valid_intrinsics_region_from       and \
       distance is None:

        # This is a common special case. This branch works identically to the
        # other path (the other side of this "if" can always be used instead),
        # but the opencv-specific code is optimized and at one point ran faster
        # than the code on the other side.
        #
        # The mask_valid_intrinsics_region_from isn't implemented in this path.
        # It COULD be, then this faster path could be used
        import cv2

        fxy_from = intrinsics_data_from[0:2]
        cxy_from = intrinsics_data_from[2:4]
        cameraMatrix_from = np.array(((fxy_from[0],          0, cxy_from[0]),
                                      ( 0,         fxy_from[1], cxy_from[1]),
                                      ( 0,                   0,           1)))

        fxy_to = intrinsics_data_to[0:2]
        cxy_to = intrinsics_data_to[2:4]
        cameraMatrix_to = np.array(((fxy_to[0],        0, cxy_to[0]),
                                    ( 0,       fxy_to[1], cxy_to[1]),
                                    ( 0,               0,         1)))

        output_shape = model_to.imagersize()
        distortion_coeffs = intrinsics_data_from[4: ]

        if Rt_to_from is not None:
            R_to_from = Rt_to_from[:3,:]
            if np.trace(R_to_from) > 3. - 1e-12:
                R_to_from = None # identity, so I pass None
        else:
            R_to_from = None

        return nps.glue( *[ nps.dummy(arr,-1) for arr in \
                            cv2.initUndistortRectifyMap(cameraMatrix_from, distortion_coeffs,
                                                        R_to_from,
                                                        cameraMatrix_to, tuple(output_shape),
                                                        cv2.CV_32FC1)],
                         axis = -1)

    W_from,H_from = model_from.imagersize()
    W_to,  H_to   = model_to.  imagersize()

    # shape: (Nheight,Nwidth,2). Contains (x,y) rows
    grid = np.ascontiguousarray(nps.mv(nps.cat(*np.meshgrid(np.arange(W_to),
                                                            np.arange(H_to))),
                                       0,-1),
                                dtype = float)
    v = mrcal.unproject(grid, lensmodel_to, intrinsics_data_to)

    if plane_n is not None:

        R_to_from = Rt_to_from[:3,:]
        t_to_from = Rt_to_from[ 3,:]

        # The homography definition. Derived in many places. For instance in
        # "Motion and structure from motion in a piecewise planar environment"
        # by Olivier Faugeras, F. Lustman.
        A_to_from = plane_d * R_to_from + nps.outer(t_to_from, plane_n)
        A_from_to = np.linalg.inv(A_to_from)
        v = nps.matmult( v, nps.transpose(A_from_to) )

    else:
        if Rt_to_from is not None:
            if distance is not None:
                v = mrcal.transform_point_Rt(mrcal.invert_Rt(Rt_to_from),
                                             v/nps.dummy(nps.mag(v),-1) * distance)
            else:
                R_to_from = Rt_to_from[:3,:]
                v = nps.matmult(v, R_to_from)

    mapxy = mrcal.project( v, lensmodel_from, intrinsics_data_from )

    if mask_valid_intrinsics_region_from:

        # Using matplotlib to compute the out-of-bounds points. It doesn't
        # support broadcasting, so I do that manually with a clump/reshape
        from matplotlib.path import Path
        region = Path(model_from.valid_intrinsics_region())
        is_inside = region.contains_points(nps.clump(mapxy,n=2)).reshape(mapxy.shape[:2])
        mapxy[ ~is_inside, :] = -1


    return mapxy.astype(np.float32)


def transform_image(image, mapxy,
                    *,
                    out = None,
                    borderMode    = None,
                    borderValue   = 0,
                    interpolation = None):

    r'''Transforms a given image using a given map

SYNOPSIS

    model_orig = mrcal.cameramodel("xxx.cameramodel")
    image_orig = mrcal.load_image("image.jpg")

    model_pinhole = mrcal.pinhole_model_for_reprojection(model_orig,
                                                         fit = "corners")

    mapxy = mrcal.image_transformation_map(model_orig, model_pinhole,
                                           intrinsics_only = True)

    image_undistorted = mrcal.transform_image(image_orig, mapxy)

    # image_undistorted is now a pinhole-reprojected version of image_orig

Given an array of pixel mappings this function can be used to transform one
image to another. If we want to convert a scene image observed by one camera
model to the image of the same scene using a different model, we can produce a
suitable transformation map with mrcal.image_transformation_map(). An example of
this common usage appears above in the synopsis.

At this time this function is a thin wrapper around cv2.remap()

ARGUMENTS

- image: a numpy array containing an image we're transforming. May be grayscale:
  shape (Nheight_input, Nwidth_input) or RGB: shape (Nheight_input,
  Nwidth_input, 3)

- mapxy: a numpy array of shape (Nheight,Nwidth,2) where Nheight and Nwidth
  represent the dimensions of the target image. This array is expected to have
  dtype=np.float32, since the internals of this function are provided by
  cv2.remap()

- out: optional numpy array of shape (Nheight,Nwidth) or (Nheight,Nwidth,3) to
  receive the result. If omitted, a new array is allocated and returned.

- borderMode: optional constant defining out-of-bounds behavior. Defaults to
  cv2.BORDER_TRANSPARENT, and is passed directly to cv2.remap(). Please see the
  docs for that function for details. This option may disappear if mrcal stops
  relying on opencv

- interpolation: optional constant defining pixel interpolation behavior.
  Defaults to cv2.INTER_LINEAR, and is passed directly to cv2.remap(). Please
  see the docs for that function for details. This option may disappear if mrcal
  stops relying on opencv

RETURNED VALUE

A numpy array of shape (Nheight, Nwidth) if grayscale or (Nheight, Nwidth, 3) if
RGB. Contains the transformed image.

    '''

    # necessary to avoid opencv crashing
    if not isinstance(image, np.ndarray): raise Exception("'image' must be a numpy array")
    if not isinstance(mapxy, np.ndarray): raise Exception("'mapxy' must be a numpy array")

    import cv2
    if borderMode is None:
        borderMode = cv2.BORDER_CONSTANT
    if interpolation is None:
        interpolation = cv2.INTER_LINEAR
    return cv2.remap(image, mapxy, None,
                     borderMode    = borderMode,
                     interpolation = interpolation,
                     borderValue   = borderValue,
                     dst           = out)



#!/usr/bin/env python3

# Copyright (c) 2017-2023 California Institute of Technology ("Caltech"). U.S.
# Government sponsorship acknowledged. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

r'''Validate deriving the input noise from the solve residuals

SYNOPSIS

  $ validate-input-noise.py [01].cameramodel

  Noise ratios measured/actual. Should be ~ 1.0
    observed, by looking at the distribution of residuals: 0.998
    predicted, by correcting the above by sqrt(1-Nstates/Nmeasurements_observed): 1.000


'''

import sys
import argparse
import re
import os

def parse_args():

    parser = \
        argparse.ArgumentParser(description = __doc__,
                                formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--observed-pixel-uncertainty',
                        type = float,
                        default = 0.3,
                        help='''How much noise to inject into perfect solves''')
    parser.add_argument('models',
                        type=str,
                        nargs='+',
                        help='''The camera models to process. Each is handled
                        individually''')

    return parser.parse_args()

args = parse_args()



# I import the LOCAL mrcal
sys.path[:0] = f"{os.path.dirname(os.path.realpath(__file__))}/..",


import mrcal
import mrcal.model_analysis
import numpy as np
import numpysane as nps


models = [mrcal.cameramodel(f) for f in args.models]

for model in models:

    optimization_inputs = model.optimization_inputs()
    mrcal.make_perfect_observations(optimization_inputs,
                                    observed_pixel_uncertainty = args.observed_pixel_uncertainty)

    mrcal.optimize(**optimization_inputs)

    # The ratio of observed noise to what I expected. Should be ~ 1.0
    noise_observed_ratio = \
        mrcal.model_analysis._observed_pixel_uncertainty_from_inputs(optimization_inputs) / args.observed_pixel_uncertainty

    Nstates       = mrcal.num_states(**optimization_inputs)
    Nmeasurements = mrcal.num_measurements(**optimization_inputs)


    # This correction is documented here:
    #   https://mrcal.secretsauce.net/docs-3.0/formulation.html#estimating-input-noise
    # This probably should be added to
    # _observed_pixel_uncertainty_from_inputs(). Today (2025/11/15) it has not
    # yet been. Because very're usually VERY overdetermined, which can be
    # validated by this script. I will add this factor later, if I discover that this is necessary
    f = np.sqrt(1 - Nstates/Nmeasurements)
    noise_predicted_ratio = noise_observed_ratio/f

    print("Noise ratios measured/actual. Should be ~ 1.0")
    print(f"  observed, by looking at the distribution of residuals: {noise_observed_ratio:.3f}")
    print(f"  predicted, by correcting the above by sqrt(1-Nstates/Nmeasurements_observed): {noise_predicted_ratio:.3f}")



#!/usr/bin/env python3

r'''Test of the mrcal-convert-lensmodel tool

'''

import sys
import numpy as np
import numpysane as nps
import os
import subprocess
import re
import io

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils


def check(filename_from, model_from,
          lensmodel_to,
          args, *,
          must_warn_of_aphysical_translation,
          what,
          distance = 3):
    process = \
        subprocess.Popen( (f"{testdir}/../mrcal-convert-lensmodel",
                           *args,
                           lensmodel_to),
                          encoding = 'ascii',
                          stdin    = subprocess.PIPE,
                          stdout   = subprocess.PIPE,
                          stderr   = subprocess.PIPE)

    with open(filename_from, "r") as f:
        text_from = f.read()
    stdout,stderr = process.communicate(input = text_from)

    if process.returncode != 0:
        testutils.confirm( False, msg = f"{what}: convert failed: {stderr}" )
    else:
        testutils.confirm( True, msg = f"{what}: convert succeeded" )

        have_warn_of_aphysical_translation = \
            re.search("WARNING: fitted camera moved by.*aphysically high", stderr)
        if must_warn_of_aphysical_translation:
            testutils.confirm( have_warn_of_aphysical_translation,
                               msg = "Expected a warning about an aphysical translation")
        else:
            testutils.confirm( not have_warn_of_aphysical_translation,
                               msg = "Expected no warning about an aphysical translation")

        with io.StringIO(stdout) as f:
            model_converted = mrcal.cameramodel(f)

        difflen, diff, q0, implied_Rt10 = \
            mrcal.projection_diff( (model_converted, model_from),
                                   use_uncertainties = False,
                                   distance          = distance,
                                   focus_radius      = 800)
        icenter = np.array(difflen.shape) // 2
        ithird  = np.array(difflen.shape) // 3

        testutils.confirm_equal( 0, difflen[icenter[0],icenter[1]],
                                 eps = 2.0,
                                 msg = f"{what}: low-enough diff at the center")

        testutils.confirm_equal( 0, difflen[ithird[0],ithird[1]],
                                 eps = 3.0,
                                 msg = f"{what}: low-enough diff at 1/3 from top-left")

filename_from = f"{testdir}/data/cam0.splined.cameramodel"
model_from    = mrcal.cameramodel(filename_from)

check( filename_from, model_from,
       "LENSMODEL_CAHVOR",
       ("--radius", "1000",
        "--intrinsics-only",
        "--sampled",
        # need multiple attempts to hit the accuracy targets below
        "--num-trials", "8",),
       must_warn_of_aphysical_translation = False,
       what = 'CAHVOR, sampled, intrinsics-only',)

check( filename_from, model_from,
       "LENSMODEL_CAHVOR",
       ("--radius", "1000",
        "--sampled",
        "--distance", "3",
        # need multiple attempts to hit the accuracy targets below
        "--num-trials", "8",),
       must_warn_of_aphysical_translation = True,
       what = 'CAHVOR, sampled at 3m',)
check( filename_from, model_from,
       "LENSMODEL_CAHVOR",
       ("--radius", "1000",
        "--sampled",
        "--distance", "3000",
        # need multiple attempts to hit the accuracy targets below
        "--num-trials", "8",),
       must_warn_of_aphysical_translation = True,
       what = 'CAHVOR, sampled at 3000m',)
check( filename_from, model_from,
       "LENSMODEL_CAHVOR",
       ("--radius", "1000",
        "--sampled",
        "--distance", "3000,3",
        # need multiple attempts to hit the accuracy targets below
        "--num-trials", "8",),
       must_warn_of_aphysical_translation = False,
       what = 'CAHVOR, sampled at 3000m,3m',)

# Need a model with optimization_inputs to do non-sampled fits
if not os.path.isdir(f"{testdir}/../../mrcal-doc-external"):
    testutils.print_blue(f"../mrcal-doc-external isn't on this disk. Skipping non-sampled tests")
else:

    filename_from = f"{testdir}/../../mrcal-doc-external/2022-11-05--dtla-overpass--samyang--alpha7/2-f22-infinity/splined.cameramodel"
    model_from = mrcal.cameramodel(filename_from)

    check( filename_from, model_from,
           "LENSMODEL_OPENCV8",
           ("--radius", "-1000"),
           must_warn_of_aphysical_translation = False,
           what = 'non-sampled fit to opencv8',)

testutils.finish()



#!/usr/bin/env python3
import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils

model_filename = f"{testdir}/data/cam0.opencv8.cameramodel"
model          = mrcal.cameramodel(model_filename)

error = None
try:
    model_loaded = mrcal._mrcal._test_python_cameramodel_converter(model_filename)
except Exception as e:
    model_loaded = None
    error = e
if model_loaded is None:
    testutils.confirm(False,
                      msg = f"Failed to load '{model_filename}': '{error}'")
else:
    testutils.confirm_equal(model_loaded['lensmodel'], model.intrinsics()[0],
                            msg = f"Loaded (from file) the correct lensmodel type")
    testutils.confirm_equal(model_loaded['intrinsics'], model.intrinsics()[1],
                            msg = f"Loaded (from file) the correct intrinsics")
    testutils.confirm_equal(model_loaded['imagersize'], model.imagersize(),
                            msg = f"Loaded (from file) the imagersize")
    testutils.confirm_equal(model_loaded['rt_cam_ref'], model.rt_cam_ref(),
                            msg = f"Loaded (from file) the correct rt_cam_ref")


error = None
try:
    model_loaded = mrcal._mrcal._test_python_cameramodel_converter(model)
except Exception as e:
    model_loaded = None
    error = e
if model_loaded is None:
    testutils.confirm(False,
                      msg = f"Failed to load cameramodel('{model_filename}'): '{error}'")
else:
    testutils.confirm_equal(model_loaded['lensmodel'], model.intrinsics()[0],
                            msg = f"Loaded (from object) the correct lensmodel type")
    testutils.confirm_equal(model_loaded['intrinsics'], model.intrinsics()[1],
                            msg = f"Loaded (from object) the correct intrinsics")
    testutils.confirm_equal(model_loaded['imagersize'], model.imagersize(),
                            msg = f"Loaded (from object) the imagersize")
    testutils.confirm_equal(model_loaded['rt_cam_ref'], model.rt_cam_ref(),
                            msg = f"Loaded (from object) the correct rt_cam_ref")


testutils.finish()



#!/usr/bin/python3

# Copyright (c) 2017-2023 California Institute of Technology ("Caltech"). U.S.
# Government sponsorship acknowledged. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

'''Routines to manipulate poses, transformations and points

Most of these are Python wrappers around the written-in-C Python extension
module mrcal._poseutils_npsp. Most of the time you want to use this module
instead of touching mrcal._poseutils_npsp directly.

All functions are exported into the mrcal module. So you can call these via
mrcal.poseutils.fff() or mrcal.fff(). The latter is preferred.

'''

import numpy as np
import numpysane as nps

# for python3
from functools import reduce

from . import _poseutils_npsp
from . import _poseutils_scipy

def r_from_R(R, *, get_gradients=False, out=None):
    r"""Compute a Rodrigues vector from a rotation matrix

SYNOPSIS

    r = mrcal.r_from_R(R)

    rotation_magnitude = nps.mag(r)
    rotation_axis      = r / rotation_magnitude

Given a rotation specified as a (3,3) rotation matrix, converts it to a
Rodrigues vector, a unit rotation axis scaled by the rotation magnitude, in
radians.

By default this function returns the Rodrigues vector(s) only. If we also want
gradients, pass get_gradients=True. Logic:

    if not get_gradients: return r
    else:                 return (r, dr/dR)

This function supports broadcasting fully.

Note that the rotation r is not unique: the rotation angle (the magnitude of r)
is only unique modulo 2 pi. And even then, rotations around axis v with angle
pi+th are equivalent to rotations around axis -v with angle pi-th. So usually we
will return r with mag(r) in [0,pi]. If the rotation angle is near pi, the
returned r might have a magnitude slightly larger than pi.

Furthermore, the gradient dr/dR needs a note. R is a (3,3) matrix, but not all
(3,3) matrices are valid rotations. Let's define vecR as a flattened (9,) vector
of all the values in R. The local linearization

  delta_r = dr/dvecR delta_vecR

is only meaningful if delta_vecR is in the valid-rotation subspace. Thus the
gradient returned by this function is any dr/dvecR + K where K maps
out-of-valid-R directions to 0. The matrix returned by this function CANNOT be
assumed to have K = 0.

ARGUMENTS

- R: array of shape (3,3). This matrix defines the rotation. It is assumed that
  this is a valid rotation (matmult(R,transpose(R)) = I, det(R) = 1), but that
  is not checked

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of Rodrigues vectors. Otherwise we return a tuple of arrays of Rodrigues
  vectors and their gradients.

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return an array of Rodrigues vector(s). Each
broadcasted slice has shape (3,)

If get_gradients: we return a tuple of arrays containing the Rodrigues vector(s)
and the gradients (r, dr/dR)

1. The Rodrigues vector. Each broadcasted slice has shape (3,)

2. The gradient dr/dR. Each broadcasted slice has shape (3,3,3). The first
   dimension selects the element of r, and the last two dimensions select the
   element of R

    """

    if get_gradients:
        return _poseutils_npsp._r_from_R_withgrad(R, out=out)
    return _poseutils_npsp._r_from_R(R, out=out)

def R_from_r(r, *, get_gradients=False, out=None):
    r"""Compute a rotation matrix from a Rodrigues vector

SYNOPSIS

    r = rotation_axis * rotation_magnitude
    R = mrcal.R_from_r(r)

Given a rotation specified as a Rodrigues vector (a unit rotation axis scaled by
the rotation magnitude, in radians.), converts it to a rotation matrix.

By default this function returns the rotation matrices only. If we also want
gradients, pass get_gradients=True. Logic:

    if not get_gradients: return R
    else:                 return (R, dR/dr)

This function supports broadcasting fully.

ARGUMENTS

- r: array of shape (3,). The Rodrigues vector that defines the rotation. This is
  a unit rotation axis scaled by the rotation magnitude, in radians

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of rotation matrices. Otherwise we return a tuple of arrays of rotation
  matrices and their gradients.

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return an array of rotation matrices. Each broadcasted
slice has shape (3,3)

If get_gradients: we return a tuple of arrays containing the rotation matrices
and the gradient (R, dR/dr):

1. The rotation matrix. Each broadcasted slice has shape (3,3). This is a valid
   rotation: matmult(R,transpose(R)) = I, det(R) = 1

2. The gradient dR/dr. Each broadcasted slice has shape (3,3,3). The first two
   dimensions select the element of R, and the last dimension selects the
   element of r

    """

    if get_gradients:
        return _poseutils_npsp._R_from_r_withgrad(r, out=out)
    return _poseutils_npsp._R_from_r(r, out=out)

def invert_R(R, *, out=None):
    r"""Invert a rotation matrix

SYNOPSIS

    print(R.shape)
    ===>
    (3,3)

    R10 = mrcal.invert_R(R01)

    print(x1.shape)
    ===>
    (3,)

    x0 = mrcal.rotate_point_R(R01, x1)

    print( nps.norm2( x1 - \
                      mrcal.rotate_point_R(R10, x0) ))
    ===>
    0

Given a rotation specified as a (3,3) rotation matrix outputs another rotation
matrix that has the opposite effect. This is simply a matrix transpose.

This function supports broadcasting fully.

In-place operation is supported; the output array may be the same as the input
array to overwrite the input.

ARGUMENTS

- R: array of shape (3,3), a rotation matrix. It is assumed that this is a valid
  rotation (matmult(R,transpose(R)) = I, det(R) = 1), but that is not checked

- out: optional argument specifying the destination. By default, a new numpy
  array is created and returned. To write the results into an existing (and
  possibly non-contiguous) array, specify it with the 'out' kwarg. If 'out' is
  given, we return the 'out' that was passed in. This is the standard behavior
  provided by numpysane_pywrap.

RETURNED VALUE

The inverse rotation matrix in an array of shape (3,3).

    """
    return _poseutils_npsp._invert_R(R, out=out)

def rt_from_Rt(Rt, *, get_gradients=False, out=None):
    r"""Compute an rt transformation from a Rt transformation

SYNOPSIS

    Rt = nps.glue(rotation_matrix,translation, axis=-2)

    print(Rt.shape)
    ===>
    (4,3)

    rt = mrcal.rt_from_Rt(Rt)

    print(rt.shape)
    ===>
    (6,)

    translation        = rt[3:]
    rotation_magnitude = nps.mag(rt[:3])
    rotation_axis      = rt[:3] / rotation_magnitude

Converts an Rt transformation to an rt transformation. Both specify a rotation
and translation. An Rt transformation is a (4,3) array formed by nps.glue(R,t,
axis=-2) where R is a (3,3) rotation matrix and t is a (3,) translation vector.
An rt transformation is a (6,) array formed by nps.glue(r,t, axis=-1) where r is
a (3,) Rodrigues vector and t is a (3,) translation vector.

Applied to a point x the transformed result is rotate(x)+t. Given a matrix R,
the rotation is defined by a matrix multiplication. x and t are stored as a row
vector (that's how numpy stores 1-dimensional arrays), but the multiplication
works as if x was a column vector (to match linear algebra conventions). See the
docs for mrcal._transform_point_Rt() for more detail.

By default this function returns the rt transformations only. If we also want
gradients, pass get_gradients=True. Logic:

    if not get_gradients: return rt
    else:                 return (rt, dr/dR)

Note that the translation gradient isn't returned: it is always the identity

This function supports broadcasting fully.

Some details about the returned r and gradients of dR are described in the
documentation for r_from_R()

ARGUMENTS

- Rt: array of shape (4,3). This matrix defines the transformation. Rt[:3,:] is
  a rotation matrix; Rt[3,:] is a translation. It is assumed that the rotation
  matrix is a valid rotation (matmult(R,transpose(R)) = I, det(R) = 1), but that
  is not checked

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of rt transformations. Otherwise we return a tuple of arrays of rt
  transformations and their gradients.

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return the rt transformation. Each broadcasted slice
has shape (6,). rt[:3] is a rotation defined as a Rodrigues vector; rt[3:] is a
translation.

If get_gradients: we return a tuple of arrays containing the rt transformation
and the gradient (rt, dr/dR):

1. The rt transformation. Each broadcasted slice has shape (6,)

2. The gradient dr/dR. Each broadcasted slice has shape (3,3,3). The first
   dimension selects the element of r, and the last two dimension select the
   element of R

    """
    if get_gradients:
        return _poseutils_npsp._rt_from_Rt_withgrad(Rt, out=out)
    return _poseutils_npsp._rt_from_Rt(Rt, out=out)

def Rt_from_rt(rt, *, get_gradients=False, out=None):
    r"""Compute an Rt transformation from a rt transformation

SYNOPSIS

    r  = rotation_axis * rotation_magnitude
    rt = nps.glue(r,t, axis=-1)

    print(rt.shape)
    ===>
    (6,)

    Rt = mrcal.Rt_from_rt(rt)

    print(Rt.shape)
    ===>
    (4,3)

    translation     = Rt[3,:]
    rotation_matrix = Rt[:3,:]

Converts an rt transformation to an Rt transformation. Both specify a rotation
and translation. An Rt transformation is a (4,3) array formed by nps.glue(R,t,
axis=-2) where R is a (3,3) rotation matrix and t is a (3,) translation vector.
An rt transformation is a (6,) array formed by nps.glue(r,t, axis=-1) where r is
a (3,) Rodrigues vector and t is a (3,) translation vector.

Applied to a point x the transformed result is rotate(x)+t. Given a matrix R,
the rotation is defined by a matrix multiplication. x and t are stored as a row
vector (that's how numpy stores 1-dimensional arrays), but the multiplication
works as if x was a column vector (to match linear algebra conventions). See the
docs for mrcal._transform_point_Rt() for more detail.

By default this function returns the Rt transformations only. If we also want
gradients, pass get_gradients=True. Logic:

    if not get_gradients: return Rt
    else:                 return (Rt, dR/dr)

Note that the translation gradient isn't returned: it is always the identity

This function supports broadcasting fully.

ARGUMENTS

- rt: array of shape (6,). This vector defines the input transformation. rt[:3]
  is a rotation defined as a Rodrigues vector; rt[3:] is a translation.

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of Rt transformations. Otherwise we return a tuple of arrays of Rt
  transformations and their gradients.

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return the Rt transformation. Each broadcasted slice
has shape (4,3). Rt[:3,:] is a rotation matrix; Rt[3,:] is a translation. The
matrix R is a valid rotation: matmult(R,transpose(R)) = I and det(R) = 1

If get_gradients: we return a tuple of arrays containing the Rt transformation
and the gradient (Rt, dR/dr):

1. The Rt transformation. Each broadcasted slice has shape (4,3,)

2. The gradient dR/dr. Each broadcasted slice has shape (3,3,3). The first two
   dimensions select the element of R, and the last dimension selects the
   element of r

    """
    if get_gradients:
        return _poseutils_npsp._Rt_from_rt_withgrad(rt, out=out)
    return _poseutils_npsp._Rt_from_rt(rt, out=out)

def invert_Rt(Rt, *, out=None):
    r"""Invert an Rt transformation

SYNOPSIS

    Rt01 = nps.glue(rotation_matrix,translation, axis=-2)

    print(Rt01.shape)
    ===>
    (4,3)

    Rt10 = mrcal.invert_Rt(Rt01)

    print(x1.shape)
    ===>
    (3,)

    x0 = mrcal.transform_point_Rt(Rt01, x1)

    print( nps.norm2( x1 - \
                      mrcal.transform_point_Rt(Rt10, x0) ))
    ===>
    0

Given an Rt transformation to convert a point representated in coordinate system
1 to coordinate system 0 (let's call it Rt01), returns a transformation that
does the reverse: converts a representation in coordinate system 0 to coordinate
system 1 (let's call it Rt10).

Thus if you have a point in coordinate system 1 (let's call it x1), we can
convert it to a representation in system 0, and then back. And we'll get the
same thing out:

  x1 == mrcal.transform_point_Rt( mrcal.invert_Rt(Rt01),
          mrcal.transform_point_Rt( Rt01, x1 ))

An Rt transformation represents a rotation and a translation. It is a (4,3)
array formed by nps.glue(R,t, axis=-2) where R is a (3,3) rotation matrix and t
is a (3,) translation vector.

Applied to a point x the transformed result is rotate(x)+t. Given a matrix R,
the rotation is defined by a matrix multiplication. x and t are stored as a row
vector (that's how numpy stores 1-dimensional arrays), but the multiplication
works as if x was a column vector (to match linear algebra conventions). See the
docs for mrcal._transform_point_Rt() for more detail.

This function supports broadcasting fully.

In-place operation is supported; the output array may be the same as the input
array to overwrite the input.

ARGUMENTS

- Rt: array of shape (4,3). This matrix defines the transformation. Rt[:3,:] is
  a rotation matrix; Rt[3,:] is a translation. It is assumed that the rotation
  matrix is a valid rotation (matmult(R,transpose(R)) = I, det(R) = 1), but that
  is not checked

- out: optional argument specifying the destination. By default, a new numpy
  array is created and returned. To write the results into an existing (and
  possibly non-contiguous) array, specify it with the 'out' kwarg. If 'out' is
  given, we return the 'out' that was passed in. This is the standard behavior
  provided by numpysane_pywrap.

RETURNED VALUE

The inverse Rt transformation in an array of shape (4,3).

    """
    return _poseutils_npsp._invert_Rt(Rt, out=out)

def invert_rt(rt, *, get_gradients=False, out=None):
    r"""Invert an rt transformation

SYNOPSIS

    r    = rotation_axis * rotation_magnitude
    rt01 = nps.glue(r,t, axis=-1)

    print(rt01.shape)
    ===>
    (6,)

    rt10 = mrcal.invert_rt(rt01)

    print(x1.shape)
    ===>
    (3,)

    x0 = mrcal.transform_point_rt(rt01, x1)

    print( nps.norm2( x1 -
                      mrcal.transform_point_rt(rt10, x0) ))
    ===>
    0

Given an rt transformation to convert a point representated in coordinate system
1 to coordinate system 0 (let's call it rt01), returns a transformation that
does the reverse: converts a representation in coordinate system 0 to coordinate
system 1 (let's call it rt10).

Thus if you have a point in coordinate system 1 (let's call it x1), we can
convert it to a representation in system 0, and then back. And we'll get the
same thing out:

  x1 == mrcal.transform_point_rt( mrcal.invert_rt(rt01),
          mrcal.transform_point_rt( rt01, x1 ))

An rt transformation represents a rotation and a translation. It is a (6,) array
formed by nps.glue(r,t, axis=-1) where r is a (3,) Rodrigues vector and t is a
(3,) translation vector.

Applied to a point x the transformed result is rotate(x)+t. x and t are stored
as a row vector (that's how numpy stores 1-dimensional arrays). See the docs for
mrcal._transform_point_rt() for more detail.

By default this function returns the rt transformation only. If we also want
gradients, pass get_gradients=True. Logic:

    if not get_gradients: return rt
    else:                 return (rt, drtout_drtin)

Note that the poseutils C API returns only

- dtout_drin
- dtout_dtin

because

- drout_drin is always -I
- drout_dtin is always 0

This Python function, however fills in those constants to return the full (and
more convenient) arrays.

This function supports broadcasting fully.

In-place operation is supported; the output array may be the same as the input
array to overwrite the input.

ARGUMENTS

- rt: array of shape (6,). This vector defines the input transformation. rt[:3]
  is a rotation defined as a Rodrigues vector; rt[3:] is a translation.

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of rt translation. Otherwise we return a tuple of arrays of rt
  translations and their gradients.

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return an array of rt transformation(s). Each
broadcasted slice has shape (6,)

If get_gradients: we return a tuple of arrays containing the rt transformation(s)
and the gradients (rt, drtout/drtin)

1. The rt transformation. Each broadcasted slice has shape (6,)

2. The gradient drtout/drtin. Each broadcasted slice has shape (6,6). The first
   dimension selects elements of rtout, and the last dimension selects elements
   of rtin

    """
    if get_gradients:
        return _poseutils_npsp._invert_rt_withgrad(rt, out=out)
    return _poseutils_npsp._invert_rt(rt, out=out)

def compose_Rt(*Rt, out=None, inverted0=False, inverted1=False):
    r"""Compose Rt transformations

SYNOPSIS

    Rt10 = nps.glue(rotation_matrix10,translation10, axis=-2)
    Rt21 = nps.glue(rotation_matrix21,translation21, axis=-2)
    Rt32 = nps.glue(rotation_matrix32,translation32, axis=-2)

    print(Rt10.shape)
    ===>
    (4,3)

    Rt30 = mrcal.compose_Rt( Rt32, Rt21, Rt10 )

    print(x0.shape)
    ===>
    (3,)

    print( nps.norm2( mrcal.transform_point_Rt(Rt30, x0) -
                      mrcal.transform_point_Rt(Rt32,
                        mrcal.transform_point_Rt(Rt21,
                          mrcal.transform_point_Rt(Rt10, x0)))))
    ===>
    0

Given 2 or more Rt transformations, returns their composition. An Rt
transformation is a (4,3) array formed by nps.glue(R,t, axis=-2) where R is a
(3,3) rotation matrix and t is a (3,) translation vector. This transformation is
defined by a matrix multiplication and an addition. x and t are stored as a row
vector (that's how numpy stores 1-dimensional arrays), but the multiplication
works as if x was a column vector (to match linear algebra conventions):

    transform_point_Rt(Rt, x) = transpose( matmult(Rt[:3,:], transpose(x)) +
                                           transpose(Rt[3,:]) ) =
                              = matmult(x, transpose(Rt[:3,:])) +
                                Rt[3,:]

This function supports broadcasting fully, so we can compose lots of
transformations at the same time.

In-place operation is supported; the output array may be the same as either of
the input arrays to overwrite the input.

ARGUMENTS

- *Rt: a list of transformations to compose. Usually we'll be composing two
  transformations, but any number could be given here. Each broadcasted slice
  has shape (4,3).

- inverted0,inverted1: optional booleans, defaulting to False. If True, the
  opposite transform is used for Rt0 and/or Rt1 respectively. inverted=True is
  only supported when exactly two transforms are given

- out: optional argument specifying the destination. By default, a new numpy
  array is created and returned. To write the results into an existing (and
  possibly non-contiguous) array, specify it with the 'out' kwarg. If 'out' is
  given, we return the 'out' that was passed in. This is the standard behavior
  provided by numpysane_pywrap.

RETURNED VALUE

An array of composed Rt transformations. Each broadcasted slice has shape (4,3)

    """
    if len(Rt) == 2:
        return _poseutils_npsp._compose_Rt(*Rt, out=out, inverted0=inverted0, inverted1=inverted1)

    if inverted0 or inverted1:
        raise Exception("compose_Rt(..., inverted...=True) is supported only if exactly 2 inputs are given")

    Rt1onwards = reduce( _poseutils_npsp._compose_Rt, Rt[1:] )
    return _poseutils_npsp._compose_Rt(Rt[0], Rt1onwards, out=out)

def compose_r(*r, get_gradients=False, out=None, inverted0=False, inverted1=False):
    r"""Compose angle-axis rotations

SYNOPSIS

    r10 = rotation_axis10 * rotation_magnitude10
    r21 = rotation_axis21 * rotation_magnitude21
    r32 = rotation_axis32 * rotation_magnitude32

    print(r10.shape)
    ===>
    (3,)

    r30 = mrcal.compose_r( r32, r21, r10 )

    print(x0.shape)
    ===>
    (3,)

    print( nps.norm2( mrcal.rotate_point_r(r30, x0) -
                      mrcal.rotate_point_r(r32,
                        mrcal.rotate_point_r(r21,
                          mrcal.rotate_point_r(r10, x0)))))
    ===>
    0

    print( [arr.shape for arr in mrcal.compose_r(r21,r10,
                                                 get_gradients = True)] )
    ===>
    [(3,), (3,3), (3,3)]

Given 2 or more axis-angle rotations, returns their composition. By default this
function returns the composed rotation only. If we also want gradients, pass
get_gradients=True. This is supported ONLY if we have EXACTLY 2 rotations to
compose. Logic:

    if not get_gradients: return r=compose(r0,r1)
    else:                 return (r=compose(r0,r1), dr/dr0, dr/dr1)

This function supports broadcasting fully, so we can compose lots of
rotations at the same time.

In-place operation is supported; the output array may be the same as either of
the input arrays to overwrite the input.

Note that the rotation r is not unique: the rotation angle (the magnitude of r)
is only unique modulo 2 pi. And even then, rotations around axis v with angle
pi+th are equivalent to rotations around axis -v with angle pi-th. So usually we
will have r with mag(r) in [0,pi]. This function might return a rotation outside
this range.

ARGUMENTS

- *r: a list of rotations to compose. Usually we'll be composing two rotations,
  but any number could be given here. Each broadcasted slice has shape (3,)

- inverted0,inverted1: optional booleans, defaulting to False. If True, the
  opposite rotation is used for r0 and/or r1 respectively. The gradients d(r0
  r1)/dr0 and d(r0 r1)/dr1 are returned in respect to the input r0 and r1.
  inverted=True is only supported when exactly two rotations are given

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of composed rotations. Otherwise we return a tuple of arrays of composed
  rotations and their gradients. Gradient reporting is only supported when
  exactly two rotations are given

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return an array of composed rotations. Each broadcasted
slice has shape (3,)

If get_gradients: we return a tuple of arrays containing the composed rotations
and the gradients (r=compose(r0,r1), dr/dr0, dr/dr1):

1. The composed rotation. Each broadcasted slice has shape (3,)

2. The gradient dr/dr0. Each broadcasted slice has shape (3,3). The first
   dimension selects the element of r, and the last dimension selects the
   element of r0

3. The gradient dr/dr1. Each broadcasted slice has shape (3,3). The first
   dimension selects the element of r, and the last dimension selects the
   element of r1

    """

    if len(r) != 2:
        if get_gradients:
            raise Exception("compose_r(..., get_gradients=True) is supported only if exactly 2 inputs are given")
        if inverted0 or inverted1:
            raise Exception("compose_r(..., inverted...=True) is supported only if exactly 2 inputs are given")

    if get_gradients:
        return _poseutils_npsp._compose_r_withgrad(*r, out=out, inverted0=inverted0, inverted1=inverted1)

    r1onwards = reduce( _poseutils_npsp._compose_r, r[1:] )
    return _poseutils_npsp._compose_r(r[0], r1onwards, out=out, inverted0=inverted0, inverted1=inverted1)

def compose_rt(*rt, get_gradients=False, out=None, inverted0=False, inverted1=False):
    r"""Compose rt transformations

SYNOPSIS

    r10 = rotation_axis10 * rotation_magnitude10
    r21 = rotation_axis21 * rotation_magnitude21
    r32 = rotation_axis32 * rotation_magnitude32

    rt10 = nps.glue(r10,t10, axis=-1)
    rt21 = nps.glue(r21,t21, axis=-1)
    rt32 = nps.glue(r32,t32, axis=-1)

    print(rt10.shape)
    ===>
    (6,)

    rt30 = mrcal.compose_rt( rt32, rt21, rt10 )

    print(x0.shape)
    ===>
    (3,)

    print( nps.norm2( mrcal.transform_point_rt(rt30, x0) -
                      mrcal.transform_point_rt(rt32,
                        mrcal.transform_point_rt(rt21,
                          mrcal.transform_point_rt(rt10, x0)))))
    ===>
    0

    print( [arr.shape for arr in mrcal.compose_rt(rt21,rt10,
                                                  get_gradients = True)] )
    ===>
    [(6,), (6,6), (6,6)]

Given 2 or more rt transformations, returns their composition. An rt
transformation is a (6,) array formed by nps.glue(r,t, axis=-1) where r is a
(3,) Rodrigues vector and t is a (3,) translation vector. This transformation is
defined by a matrix multiplication and an addition. x and t are stored as a row
vector (that's how numpy stores 1-dimensional arrays), but the multiplication
works as if x was a column vector (to match linear algebra conventions):

    transform_point_rt(rt, x) = transpose( matmult(R_from_r(rt[:3]), transpose(x)) +
                                           transpose(rt[3,:]) ) =
                              = matmult(x, transpose(R_from_r(rt[:3]))) +
                                rt[3:]

By default this function returns the composed transformation only. If we also
want gradients, pass get_gradients=True. This is supported ONLY if we have
EXACTLY 2 transformations to compose. Logic:

    if not get_gradients: return rt=compose(rt0,rt1)
    else:                 return (rt=compose(rt0,rt1), dr/drt0, dr/drt1)

Note that the poseutils C API returns only

- dr_dr0
- dr_dr1
- dt_dr0
- dt_dt1
- dt/dr1
- dt/dt0

because

- dr/dt0 is always 0
- dr/dt1 is always 0

This Python function, however fills in those constants to return the full (and
more convenient) arrays.

This function supports broadcasting fully, so we can compose lots of
transformations at the same time.

In-place operation is supported; the output array may be the same as either of
the input arrays to overwrite the input.

Some details about the returned r are described in the documentation for
compose_r()

ARGUMENTS

- *rt: a list of transformations to compose. Usually we'll be composing two
  transformations, but any number could be given here. Each broadcasted slice
  has shape (6,)

- inverted0,inverted1: optional booleans, defaulting to False. If True, the
  opposite transform is used for rt0 and/or rt1 respectively. The gradients
  d(rt0 rt1)/drt0 and d(rt0 rt1)/drt1 are returned in respect to the input rt0
  and rt1. inverted=True is only supported when exactly two transforms are given

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of composed transformations. Otherwise we return a tuple of arrays of
  composed transformations and their gradients. Gradient reporting is only
  supported when exactly two transformations are given

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return an array of composed rt transformations. Each
broadcasted slice has shape (6,)

If get_gradients: we return a tuple of arrays containing the composed
transformations and the gradients (rt=compose(rt0,rt1),
drt/drt0,drt/drt1):

1. The composed transformation. Each broadcasted slice has shape (6,)

2. The gradient drt/drt0. Each broadcasted slice has shape (6,6). The first
   dimension selects the element of rt, and the last dimension selects the
   element of rt0

3. The gradient drt/drt1. Each broadcasted slice has shape (6,6). The first
   dimension selects the element of rt, and the last dimension selects the
   element of rt1

    """

    if len(rt) != 2:
        if get_gradients:
            raise Exception("compose_rt(..., get_gradients=True) is supported only if exactly 2 inputs are given")
        if inverted0 or inverted1:
            raise Exception("compose_rt(..., inverted...=True) is supported only if exactly 2 inputs are given")

    if get_gradients:
        return _poseutils_npsp._compose_rt_withgrad(*rt, out=out, inverted0=inverted0, inverted1=inverted1)

    rt1onwards = reduce( _poseutils_npsp._compose_rt, rt[1:] )
    return _poseutils_npsp._compose_rt(rt[0], rt1onwards, out=out, inverted0=inverted0, inverted1=inverted1)

def compose_rt_tinyrt0_gradientrt0(rt1, out=None):
    r"""Special-case composition for the uncertainty computation

SYNOPSIS

    r1  = rotation_axis1 * rotation_magnitude1
    rt1 = nps.glue(r1, t1,  axis=-1)

    drt01_drt0 = compose_rt_tinyrt0_gradientrt0(rt1)

    ### Another way to get the same thing (but possibly less efficiently)
     _,drt01_drt0,_ = compose_rt(np.zeros((6,),),
                                 rt1,
                                 get_gradients=True)

This is a special-case subset of compose_rt(). It is the same, except:

- rt0 is assumed to be 0, so we don't ingest it, and we don't report the
  composition result
- we ONLY report the drt01/drt0 gradient

This special-case function is a part of the projection uncertainty computation,
so it exists separate from compose_rt(). See the documentation for compose_rt()
for all the details.

This function supports broadcasting fully

ARGUMENTS

- rt1: the second of the two transformations being composed. The first is an
  identity, so it's not given as an argument

- out: optional argument specifying the destination. By default, a new numpy
  array is created and returned. To write the results into an existing (and
  possibly non-contiguous) array, specify it with the 'out' kwarg

RETURNED VALUE

We return a single array of shape (..., 6,6): drt01/drt0

    """

    # R0 (R1 p + t1) + t0 = R0 R1 p + (R0 t1 + t0)
    # -> R01 = R0 R1
    # -> t01 = R0 t1 + t0

    # At rt0 ~ identity we have:

    #   dt01/dr0 = d(R0 t1)/dr0

    # rotate_point_r_core() says that

    #   const val_withgrad_t<N> cross[3] =
    #       {
    #           (rg[1]*x_ing[2] - rg[2]*x_ing[1])*sign,
    #           (rg[2]*x_ing[0] - rg[0]*x_ing[2])*sign,
    #           (rg[0]*x_ing[1] - rg[1]*x_ing[0])*sign
    #       };
    #   const val_withgrad_t<N> inner =
    #       rg[0]*x_ing[0] +
    #       rg[1]*x_ing[1] +
    #       rg[2]*x_ing[2];

    #   // Small rotation. I don't want to divide by 0, so I take the limit
    #   //   lim(th->0, xrot) =
    #   //     = x + cross(r, x) + r rt x lim(th->0, (1 - cos(th)) / (th*th))
    #   //     = x + cross(r, x) + r rt x lim(th->0, sin(th) / (2*th))
    #   //     = x + cross(r, x) + r rt x/2
    #   for(int i=0; i<3; i++)
    #       x_outg[i] =
    #           x_ing[i] +
    #           cross[i] +
    #           rg[i]*inner / 2.;

    # So t01 = t0 + t1 + linear(r0) + quadratic(r0)

    # r0 ~ 0 so I ignore the quadratic term:

    #   dt01/dr0 = d(cross(r0,t1))/dr0
    #            = -d(cross(t1,r0))/dr0
    #            = -d(skew_symmetric(t1) r0))/dr0
    #            = -skew_symmetric(t1)

    # Thus

    #   drt01/drt0 = [ dr01/dr0  dr01/dt0  ] = [ dr01/dr0              0 ]
    #                [ dt01/dr0  dt01/dt0  ] = [ -skew_symmetric(t1)   I ]

    # I call a function to get dr01_dr0

    out_shape = rt1.shape + (6,)
    if out is None:
        out = np.zeros(out_shape, dtype=float)
    else:
        if not out.shape == out_shape:
            raise Exception(f"The given 'out' array has shape {out.shape} but rt1 has shape {rt1.shape}, so 'out' should have shape {out_shape}")
        out[:] = 0

    _poseutils_npsp.skew_symmetric(rt1[..., 3:],
                                   out = out[..., 3:, :3])
    out[..., 3:, :3] *= -1

    out[..., 0+3, 0+3] = 1.
    out[..., 1+3, 1+3] = 1.
    out[..., 2+3, 2+3] = 1.

    _poseutils_npsp.compose_r_tinyr0_gradientr0(rt1[..., :3],
                                                out = out[..., :3, :3])

    return out

def compose_rt_tinyrt1_gradientrt1(rt0, out=None):
    r"""Special-case composition for the uncertainty computation

SYNOPSIS

    r0  = rotation_axis0 * rotation_magnitude0
    rt0 = nps.glue(r0, t0,  axis=-1)

    drt01_drt1 = compose_rt_tinyrt1_gradientrt1(rt0)

    ### Another way to get the same thing (but possibly less efficiently)
     _,_,drt01_drt1 = compose_rt(rt0,
                                 np.zeros((6,),),
                                 get_gradients=True)

This is a special-case subset of compose_rt(). It is the same, except:

- rt1 is assumed to be 0, so we don't ingest it, and we don't report the
  composition result
- we ONLY report the drt01/drt1 gradient

This special-case function is a part of the projection uncertainty computation,
so it exists separate from compose_rt(). See the documentation for compose_rt()
for all the details.

This function supports broadcasting fully

ARGUMENTS

- rt0: the first of the two transformations being composed. The second is an
  identity, so it's not given as an argument

- out: optional argument specifying the destination. By default, a new numpy
  array is created and returned. To write the results into an existing (and
  possibly non-contiguous) array, specify it with the 'out' kwarg

RETURNED VALUE

We return a single array of shape (..., 6,6): drt01/drt1

    """

    # R0 (R1 p + t1) + t0 = R0 R1 p + (R0 t1 + t0)
    # -> R01 = R0 R1
    # -> t01 = R0 t1 + t0

    # rotate_point_r_core() says that

    #   const val_withgrad_t<N> cross[3] =
    #       {
    #           (rg[1]*x_ing[2] - rg[2]*x_ing[1])*sign,
    #           (rg[2]*x_ing[0] - rg[0]*x_ing[2])*sign,
    #           (rg[0]*x_ing[1] - rg[1]*x_ing[0])*sign
    #       };
    #   const val_withgrad_t<N> inner =
    #       rg[0]*x_ing[0] +
    #       rg[1]*x_ing[1] +
    #       rg[2]*x_ing[2];

    #   if(th2.x < 1e-10)
    #   {
    #     // Small rotation. I don't want to divide by 0, so I take the limit
    #     //   lim(th->0, xrot) =
    #     //     = x + cross(r, x) + r rt x lim(th->0, (1 - cos(th)) / (th*th))
    #     //     = x + cross(r, x) + r rt x lim(th->0, sin(th) / (2*th))
    #     //     = x + cross(r, x) + r rt x/2
    #     for(int i=0; i<3; i++)
    #         x_outg[i] =
    #             x_ing[i] +
    #             cross[i] +
    #             rg[i]*inner / 2.;
    #   }
    #   else
    #   {
    #     xrot = x cos(th) + cross(r, x)*sin(th)/th + r rt x (1 - cos(th)) / (th*th)
    #   }

    # So if r0 is small:
    #   t01      = t0 + t1 + cross(r0,t1) + r0/2 inner(r0,t1)
    #   dt01/dt1 = I + skew_symmetric(r0) + outer(r0,r0)/2
    # else:
    #   t01      = t0 + t1 cos(th) + cross(r0,t1)*sin(th)/th + outer(r0,r0) t1 (1 - cos(th)) / (th*th)
    #   dt01/dt1 = I cos(th) + skew_symmetric(r0)*sin(th)/th + outer(r0,r0) (1 - cos(th)) / (th*th)

    #   drt01/drt1 = [ dr01/dr1  dr01/dt1  ] = [ dr01/dr1   0        ]
    #                [ dt01/dr1  dt01/dt1  ] = [ 0          dt01/dt1 ]

    out_shape = rt0.shape + (6,)
    if out is None:
        out = np.zeros(out_shape, dtype=float)
    else:
        if not out.shape == out_shape:
            raise Exception(f"The given 'out' array has shape {out.shape} but rt0 has shape {rt0.shape}, so 'out' should have shape {out_shape}")
        out[:] = 0


    # dummy arrays
    rt01       = np.zeros(rt0.shape,        dtype=float)
    drt01_drt0 = np.zeros(rt0.shape + (6,), dtype=float)

    return \
        compose_rt(rt0,
                   np.zeros((6,),),
                   get_gradients=True,
                   out = (rt01,drt01_drt0,out))[2]



    _poseutils_npsp.skew_symmetric(rt0[..., :3],
                                   out = out[..., 3:, 3:])

    th2 = nps.norm2(rt0[..., :3])
    s = np.array(th2 * 0.)
    mask_r0_tiny = th2 < 1e-10
    if np.any(mask_r0_tiny):
        s[ mask_r0_tiny] = 0.5

    if np.any(~mask_r0_tiny):
        s[~mask_r0_tiny] = (1 - np.cos(np.sqrt(th2[~mask_r0_tiny])) / th2[~mask_r0_tiny])

    out[..., 3:, 3:] += np.eye(3) + nps.outer(rt0[..., :3],rt0[..., :3]) * s

    _poseutils_npsp.compose_r_tinyr1_gradientr1(rt0[..., :3],
                                                out = out[..., :3, :3])

    return out


def rotate_point_r(r, x, *, get_gradients=False, out=None, inverted=False):
    r"""Rotate point(s) using a Rodrigues vector

SYNOPSIS

    r = rotation_axis * rotation_magnitude

    print(r.shape)
    ===>
    (3,)

    print(x.shape)
    ===>
    (10,3)

    print(mrcal.rotate_point_r(r, x).shape)
    ===>
    (10,3)

    print( [arr.shape for arr in mrcal.rotate_point_r(r, x,
                                                      get_gradients = True)] )
    ===>
    [(10,3), (10,3,3), (10,3,3)]

Rotate point(s) x by a Rodrigues rotation r. By default this function returns
the rotated points only. If we also want gradients, pass get_gradients=True.
Logic:

    if not get_gradients: return u=r(x)
    else:                 return (u=r(x),du/dr,du/dx)

This function supports broadcasting fully, so we can rotate lots of points at
the same time and/or apply lots of different rotations at the same time

In-place operation is supported; the output array may be the same as the input
arrays to overwrite the input.

ARGUMENTS

- r: array of shape (3,). The Rodrigues vector that defines the rotation. This is
  a unit rotation axis scaled by the rotation magnitude, in radians

- x: array of shape (3,). The point being rotated

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of rotated points. Otherwise we return a tuple of arrays of rotated
  points and their gradients.

- inverted: optional boolean, defaulting to False. If True, the opposite
  rotation is computed. The gradient du/dr is returned in respect to the input r

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return an array of rotated point(s). Each broadcasted
slice has shape (3,)

If get_gradients: we return a tuple of arrays containing the rotated points and
the gradients (u=r(x),du/dr,du/dx):

A tuple (u=r(x),du/dr,du/dx):

1. The rotated point(s). Each broadcasted slice has shape (3,)

2. The gradient du/dr. Each broadcasted slice has shape (3,3). The first
   dimension selects the element of u, and the last dimension selects the
   element of r

3. The gradient du/dx. Each broadcasted slice has shape (3,3). The first
   dimension selects the element of u, and the last dimension selects the
   element of x

    """
    if not get_gradients:
        return _poseutils_npsp._rotate_point_r(r,x, out=out, inverted=inverted)
    return _poseutils_npsp._rotate_point_r_withgrad(r,x, out=out, inverted=inverted)

def rotate_point_R(R, x, *, get_gradients=False, out=None, inverted=False):
    r"""Rotate point(s) using a rotation matrix

SYNOPSIS

    r = rotation_axis * rotation_magnitude
    R = mrcal.R_from_r(r)

    print(R.shape)
    ===>
    (3,3)

    print(x.shape)
    ===>
    (10,3)

    print( mrcal.rotate_point_R(R, x).shape )
    ===>
    (10,3)

    print( [arr.shape for arr in mrcal.rotate_point_R(R, x,
                                                      get_gradients = True)] )
    ===>
    [(10,3), (10,3,3,3), (10,3,3)]

Rotate point(s) by a rotation matrix. This is a matrix multiplication. x is
stored as a row vector (that's how numpy stores 1-dimensional arrays), but the
multiplication works as if x was a column vector (to match linear algebra
conventions):

    rotate_point_R(R,x) = transpose( matmult(R, transpose(x))) =
                        = matmult(x, transpose(R))

By default this function returns the rotated points only. If we also want
gradients, pass get_gradients=True. Logic:

    if not get_gradients: return u=R(x)
    else:                 return (u=R(x),du/dR,du/dx)

This function supports broadcasting fully, so we can rotate lots of points at
the same time and/or apply lots of different rotations at the same time

In-place operation is supported; the output array may be the same as the input
arrays to overwrite the input.

ARGUMENTS

- R: array of shape (3,3). This matrix defines the rotation. It is assumed that
  this is a valid rotation (matmult(R,transpose(R)) = I, det(R) = 1), but that
  is not checked

- x: array of shape (3,). The point being rotated

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of rotated points. Otherwise we return a tuple of arrays of rotated
  points and their gradients.

- inverted: optional boolean, defaulting to False. If True, the opposite
  rotation is computed. The gradient du/dR is returned in respect to the input R

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return an array of rotated point(s). Each broadcasted
slice has shape (3,)

If get_gradients: we return a tuple of arrays containing the rotated points and
the gradients (u=R(x),du/dR,du/dx):

1. The rotated point(s). Each broadcasted slice has shape (3,)

2. The gradient du/dR. Each broadcasted slice has shape (3,3,3). The first
   dimension selects the element of u, and the last 2 dimensions select the
   element of R

3. The gradient du/dx. Each broadcasted slice has shape (3,3). The first
   dimension selects the element of u, and the last dimension selects the
   element of x

    """

    if not get_gradients:
        return _poseutils_npsp._rotate_point_R(R,x, out=out, inverted=inverted)
    return _poseutils_npsp._rotate_point_R_withgrad(R,x, out=out, inverted=inverted)

def transform_point_rt(rt, x, *, get_gradients=False, out=None, inverted=False):
    r"""Transform point(s) using an rt transformation

SYNOPSIS

    r  = rotation_axis * rotation_magnitude
    rt = nps.glue(r,t, axis=-1)

    print(rt.shape)
    ===>
    (6,)

    print(x.shape)
    ===>
    (10,3)

    print( mrcal.transform_point_rt(rt, x).shape )
    ===>
    (10,3)

    print( [arr.shape
            for arr in mrcal.transform_point_rt(rt, x,
                                                get_gradients = True)] )
    ===>
    [(10,3), (10,3,6), (10,3,3)]

Transform point(s) by an rt transformation: a (6,) array formed by
nps.glue(r,t, axis=-1) where r is a (3,) Rodrigues vector and t is a (3,)
translation vector. This transformation is defined by a matrix multiplication
and an addition. x and t are stored as a row vector (that's how numpy stores
1-dimensional arrays), but the multiplication works as if x was a column vector
(to match linear algebra conventions):

    transform_point_rt(rt, x) = transpose( matmult(R_from_r(rt[:3]), transpose(x)) +
                                           transpose(rt[3,:]) ) =
                              = matmult(x, transpose(R_from_r(rt[:3]))) +
                                rt[3:]

By default this function returns the transformed points only. If we also want
gradients, pass get_gradients=True. Logic:

    if not get_gradients: return u=rt(x)
    else:                 return (u=rt(x),du/drt,du/dx)

This function supports broadcasting fully, so we can transform lots of points at
the same time and/or apply lots of different transformations at the same time

In-place operation is supported; the output array may be the same as the input
arrays to overwrite the input.

ARGUMENTS

- rt: array of shape (6,). This vector defines the transformation. rt[:3] is a
  rotation defined as a Rodrigues vector; rt[3:] is a translation.

- x: array of shape (3,). The point being transformed

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of transformed points. Otherwise we return a tuple of arrays of
  transformed points and their gradients.

- inverted: optional boolean, defaulting to False. If True, the opposite
  transformation is computed. The gradient du/drt is returned in respect to the
  input rt

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return an array of transformed point(s). Each
broadcasted slice has shape (3,)

If get_gradients: we return a tuple of arrays of transformed points and the
gradients (u=rt(x),du/drt,du/dx):

1. The transformed point(s). Each broadcasted slice has shape (3,)

2. The gradient du/drt. Each broadcasted slice has shape (3,6). The first
   dimension selects the element of u, and the last dimension selects the
   element of rt

3. The gradient du/dx. Each broadcasted slice has shape (3,3). The first
   dimension selects the element of u, and the last dimension selects the
   element of x

    """

    if not get_gradients:
        return _poseutils_npsp._transform_point_rt(rt,x, out=out, inverted=inverted)
    return _poseutils_npsp._transform_point_rt_withgrad(rt,x, out=out, inverted=inverted)

def transform_point_Rt(Rt, x, *, get_gradients=False, out=None, inverted=False):
    r"""Transform point(s) using an Rt transformation

SYNOPSIS

    Rt = nps.glue(rotation_matrix,translation, axis=-2)

    print(Rt.shape)
    ===>
    (4,3)

    print(x.shape)
    ===>
    (10,3)

    print( mrcal.transform_point_Rt(Rt, x).shape )
    ===>
    (10,3)

    print( [arr.shape
            for arr in mrcal.transform_point_Rt(Rt, x,
                                                get_gradients = True)] )
    ===>
    [(10,3), (10,3,4,3), (10,3,3)]

Transform point(s) by an Rt transformation: a (4,3) array formed by
nps.glue(R,t, axis=-2) where R is a (3,3) rotation matrix and t is a (3,)
translation vector. This transformation is defined by a matrix multiplication
and an addition. x and t are stored as a row vector (that's how numpy stores
1-dimensional arrays), but the multiplication works as if x was a column vector
(to match linear algebra conventions):

    transform_point_Rt(Rt, x) = transpose( matmult(Rt[:3,:], transpose(x)) +
                                           transpose(Rt[3,:]) ) =
                              = matmult(x, transpose(Rt[:3,:])) +
                                Rt[3,:]

By default this function returns the transformed points only. If we also want
gradients, pass get_gradients=True. Logic:

    if not get_gradients: return u=Rt(x)
    else:                 return (u=Rt(x),du/dRt,du/dx)

This function supports broadcasting fully, so we can transform lots of points at
the same time and/or apply lots of different transformations at the same time

In-place operation is supported; the output array may be the same as the input
arrays to overwrite the input.

ARGUMENTS

- Rt: array of shape (4,3). This matrix defines the transformation. Rt[:3,:] is
    a rotation matrix; Rt[3,:] is a translation. It is assumed that the rotation
    matrix is a valid rotation (matmult(R,transpose(R)) = I, det(R) = 1), but that
    is not checked

- x: array of shape (3,). The point being transformed

- get_gradients: optional boolean. By default (get_gradients=False) we return an
  array of transformed points. Otherwise we return a tuple of arrays of
  transformed points and their gradients.

- inverted: optional boolean, defaulting to False. If True, the opposite
  transformation is computed. The gradient du/dRt is returned in respect to the
  input Rt

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg. If not
  get_gradients: 'out' is the one numpy array we will write into. Else: 'out' is
  a tuple of all the output numpy arrays. If 'out' is given, we return the 'out'
  that was passed in. This is the standard behavior provided by
  numpysane_pywrap.

RETURNED VALUE

If not get_gradients: we return an array of transformed point(s). Each
broadcasted slice has shape (3,)

If get_gradients: we return a tuple of arrays of transformed points and the
gradients (u=Rt(x),du/dRt,du/dx):

1. The transformed point(s). Each broadcasted slice has shape (3,)

2. The gradient du/dRt. Each broadcasted slice has shape (3,4,3). The first
   dimension selects the element of u, and the last 2 dimensions select the
   element of Rt

3. The gradient du/dx. Each broadcasted slice has shape (3,3). The first
   dimension selects the element of u, and the last dimension selects the
   element of x

    """

    if not get_gradients:
        return _poseutils_npsp._transform_point_Rt(Rt,x, out=out, inverted=inverted)
    return _poseutils_npsp._transform_point_Rt_withgrad(Rt,x, out=out, inverted=inverted)


quat_from_R = _poseutils_scipy.quat_from_R

def qt_from_Rt(Rt, *, out=None):
    r"""Compute a qt transformation from a Rt transformation

SYNOPSIS

    Rt = nps.glue(rotation_matrix,translation, axis=-2)

    print(Rt.shape)
    ===>
    (4,3)

    qt = mrcal.qt_from_Rt(Rt)

    print(qt.shape)
    ===>
    (7,)

    quat        = qt[:4]
    translation = qt[4:]

Converts an Rt transformation to a qt transformation. Both specify a rotation
and translation. An Rt transformation is a (4,3) array formed by nps.glue(R,t,
axis=-2) where R is a (3,3) rotation matrix and t is a (3,) translation vector.
A qt transformation is a (7,) array formed by nps.glue(q,t, axis=-1) where q is
a (4,) unit quaternion and t is a (3,) translation vector.

Applied to a point x the transformed result is rotate(x)+t. Given a matrix R,
the rotation is defined by a matrix multiplication. x and t are stored as a row
vector (that's how numpy stores 1-dimensional arrays), but the multiplication
works as if x was a column vector (to match linear algebra conventions). See the
docs for mrcal._transform_point_Rt() for more detail.

This function supports broadcasting fully.

Note: mrcal does not use unit quaternions anywhere to represent rotations. This
function is provided for convenience, but isn't thoroughly tested.

ARGUMENTS

- Rt: array of shape (4,3). This matrix defines the transformation. Rt[:3,:] is
  a rotation matrix; Rt[3,:] is a translation. It is assumed that the rotation
  matrix is a valid rotation (matmult(R,transpose(R)) = I, det(R) = 1), but that
  is not checked

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg.

RETURNED VALUE

We return the qt transformation. Each broadcasted slice has shape (7,). qt[:4]
is a rotation defined as a unit quaternion; qt[4:] is a translation.

    """
    if out is not None:
        qt = out
    else:
        qt = np.zeros(Rt.shape[:-2] + (7,), dtype=float)

    _poseutils_scipy.quat_from_R(Rt[..., :3, :], out=qt[..., :4])
    qt[..., 4:] = Rt[..., 3, :]
    return qt


def Rt_from_qt(qt, *, out=None):
    r"""Compute an Rt transformation from a qt transformation

SYNOPSIS

    qt = nps.glue(q,t, axis=-1)

    print(qt.shape)
    ===>
    (7,)

    Rt = mrcal.Rt_from_qt(qt)

    print(Rt.shape)
    ===>
    (4,3)

    translation     = Rt[3,:]
    rotation_matrix = Rt[:3,:]

Converts a qt transformation to an Rt transformation. Both specify a rotation
and translation. An Rt transformation is a (4,3) array formed by nps.glue(R,t,
axis=-2) where R is a (3,3) rotation matrix and t is a (3,) translation vector.
A qt transformation is a (7,) array formed by nps.glue(q,t, axis=-1) where q is
a (4,) unit quaternion and t is a (3,) translation vector.

Applied to a point x the transformed result is rotate(x)+t. Given a matrix R,
the rotation is defined by a matrix multiplication. x and t are stored as a row
vector (that's how numpy stores 1-dimensional arrays), but the multiplication
works as if x was a column vector (to match linear algebra conventions). See the
docs for mrcal._transform_point_Rt() for more detail.

This function supports broadcasting fully.

Note: mrcal does not use unit quaternions anywhere to represent rotations. This
function is provided for convenience, but isn't thoroughly tested.

ARGUMENTS

- qt: array of shape (7,). This vector defines the input transformation. qt[:4]
  is a rotation defined as a unit quaternion; qt[4:] is a translation.

- out: optional argument specifying the destination. By default, new numpy
  array(s) are created and returned. To write the results into existing (and
  possibly non-contiguous) arrays, specify them with the 'out' kwarg.

RETURNED VALUE

We return the Rt transformation. Each broadcasted slice has shape (4,3).
Rt[:3,:] is a rotation matrix; Rt[3,:] is a translation. The matrix R is a valid
rotation: matmult(R,transpose(R)) = I and det(R) = 1

    """
    if out is not None:
        Rt = out
    else:
        Rt = np.zeros(qt.shape[:-1] + (4,3), dtype=float)

    _poseutils_npsp.R_from_quat(qt[..., :4], out=Rt[..., :3, :])
    Rt[..., 3, :] = qt[..., 4:]
    return Rt



#!/usr/bin/env python3

import sys
import numpy as np
import numpysane as nps
import os
testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal

from testutils import *
from test_calibration_helpers import grad,grad__r_from_R

from test_poseutils_helpers import \
    R_from_r,                      \
    r_from_R,                      \
    compose_r

def wrap_r_unconditional(r, dr_dX = None):
    '''Unwrap a Rodrigues vector r

    returns r'     if dr_dX is None
    returns dr'_dX if dr_dX is not None

    I define rotations with a Rodrigues vector r = th v. Where v is the unit
    vector representing the rotation axis; and th is a scalar representing the
    magnitude or this rotation, in radians. Thus rotations th v and (th + 2pi*n)
    v are identical for all integer n. Furthermore the rotation (pi+th) v is
    equivalent to a rotation (pi-th) (-v). So usually we have th = mag(r) in
    [0,pi].

    For various analyses I want to convert a rotation r to its equivalent
    rotation 2*pi rad away:

      r' = v (th - 2pi)
         = r/magr (magr - 2pi)
         = r - r/magr 2pi
         = r (1 - 2pi/magr)

    Let

      k = 1 - 2pi/magr

    So

      r' = r k

    As noted above, usually magr <= pi, so k < 0. Thus

      magr' = -magr k

    Let's apply this wrapping a second time:

      r'' = r' (1 - 2pi/magr')
          = r k (1 + 2pi/magr/k)
          = r k (1 + (1-k) /k)
          = r (k + 1 - k)
          = r

    So this wrapping operation switches back/forth between two modes

    '''


    if dr_dX is None:
        if nps.mag(r) == 0:
            return r
        return r - r/nps.mag(r) * 2.*np.pi

    if nps.mag(r) == 0:
        return dr_dX

    # drw_dX = d( r - r/nps.mag(r) * 2.*np.pi )/dX
    #        = dr_dX - 2pi (dr_dX/nps.mag(r) + r d/dx (1/magr))
    #        = dr_dX - 2pi (dr_dX/nps.mag(r) - r /magr^2 /2magr 2rt dr_dX )
    #        = dr_dX - 2pi (dr_dX - r rt /norm2(r) dr_dX ) /nps.mag(r)
    #        = dr_dX + 2pi /nps.mag(r) (r rt /norm2(r) - I) dr_dX

    # dr_dX has shape (3,...) so r rt dr_dX also has shape (3,...) I clump
    # the trailing shapes so that dr_dX.shape is (3,N), and then reshape it
    # at the end
    s = dr_dX.shape[1:]
    # shape (3,N)
    dr_dX = nps.clump(dr_dX, n = -(dr_dX.ndim-1))

    drw_dX =    \
        dr_dX + \
        2.*np.pi / nps.mag(r) * nps.matmult(nps.outer(r,r)/nps.norm2(r) - np.eye(r.size),
                                            dr_dX)
    return drw_dX.reshape((3,) + s)

def wrap_r(r,
           *,
           r_match_direction = None,
           dr_dX             = None):

    '''Unwrap a Rodrigues vector r

    returns r'     if dr_dX is None
    returns dr'_dX if dr_dX is not None

    This function only wraps the argument if it needs to: if mag(r) > np.pi or
    if we're pointing opposite r_match_direction

    '''

    if r_match_direction is not None:
        if nps.inner(r, r_match_direction) > 0:
            return r if dr_dX is None else dr_dX
        return wrap_r_unconditional(r, dr_dX = dr_dX)
    if nps.mag(r) <= np.pi:
        return r if dr_dX is None else dr_dX
    return wrap_r_unconditional(r, dr_dX = dr_dX)


################### Check the behavior around the th=0, th=180, th=360
################### singularities. Gradients and values should be correct
axes = \
    np.array(((1.,  2.,  0.1),
              (1.,  0,   0),
              (0,   0,   -1.),))
axes /= nps.dummy(nps.mag(axes), -1) # normalize
for iaxis,axis in enumerate(axes):
    for th0 in (-np.pi, 0, np.pi):
        for dth in (-1e-4, -1e-10, 0, 1e-10, 1e-4):

            r = (th0 + dth) * axis

            ######### R_from_r, r_from_R
            if True:

                R,dR_dr    = mrcal.R_from_r(r, get_gradients=True)

                R_ref      = R_from_r(r)
                dR_dr__ref = grad(R_from_r,r)

                confirm_equal( R,
                               R_ref,
                               msg=f'R_from_r result near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')
                confirm_equal( dR_dr,
                               dR_dr__ref,
                               msg=f'R_from_r J near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')

                # I check R_roundtrip. The dr/dR computation assumes this
                r_roundtrip,dr_dR = mrcal.r_from_R(R,           get_gradients=True)
                R_roundtrip,dR_dr = mrcal.R_from_r(r_roundtrip, get_gradients=True)

                confirm_equal( mrcal.compose_r(r_roundtrip, -r),
                               0,
                               eps = 1e-8,
                               msg=f'roundtrip r result near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')
                confirm_equal( nps.matmult(R_roundtrip, mrcal.invert_R(R)) - np.eye(3),
                               0,
                               eps = 1e-8,
                               msg=f'roundtrip R result near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')

                dr_dR__ref = grad__r_from_R(R_ref)

                confirm_equal( dr_dR,
                               dr_dR__ref,
                               relative    = True,
                               worstcase   = True,
                               eps         = 1e-3,
                               reldiff_eps = 1e-5,
                               msg         = f'r_from_R J near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')

            ######### compose_r
            if True:

                r1_simple = np.array((-0.02, -1.2, 0.4),)

                inv_r1_simple_r = mrcal.compose_r(-r1_simple,r)

                # it isn't possible to correctly decide if we should or should
                # not wrap the result for ALL cases (some cases have mag(r)=pi
                # exactly, up to machine precision). So I try both, and pick the
                # closer one
                r_roundtrip = mrcal.compose_r( r1_simple, inv_r1_simple_r )
                r_wrapped   = wrap_r_unconditional(r)
                confirm_equal( r_roundtrip,
                               r if nps.norm2(r_roundtrip-r) < nps.norm2(r_roundtrip-r_wrapped) \
                               else r_wrapped,
                               worstcase = True,
                               eps = 1e-6,
                               msg='compose()')

                for (r0,r1) in ((r, r1_simple),
                                (r,-r),
                                (r, r),
                                (r1_simple, inv_r1_simple_r)):
                    ###### r01
                    r01, dr01_dr0, dr01_dr1 = mrcal.compose_r(r0,r1, get_gradients = True)
                    r01_ref                 = compose_r(r0,r1)
                    confirm_equal( r01,
                                   r01_ref,
                                   worstcase = True,
                                   relative  = True,
                                   eps       = 1e-3,
                                   msg=f'compose_r(r0,r1) near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')
                    dr01_dr0__ref = grad(lambda r0: compose_r(r0,r1),
                                         r0,
                                         forward_differences = True,
                                         switch = wrap_r_unconditional,
                                         step = 1e-7)
                    dr01_dr1__ref = grad(lambda r1: compose_r(r0,r1),
                                         r1,
                                         forward_differences = True,
                                         switch = wrap_r_unconditional,
                                         step = 1e-7)
                    confirm_equal( dr01_dr0,
                                   dr01_dr0__ref,
                                   worstcase = True,
                                   relative  = True,
                                   eps       = 1e-2,
                                   reldiff_eps=1e-3,
                                   msg=f'compose_r(r0,r1) dr01_dr0 near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')
                    confirm_equal( dr01_dr1,
                                   dr01_dr1__ref,
                                   worstcase = True,
                                   relative  = True,
                                   eps       = 1e-2,
                                   reldiff_eps=1e-3,
                                   msg=f'compose_r(r0,r1) dr01_dr1 near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')

                    if r0 is r1: continue

                    ###### r10
                    r10, dr10_dr1, dr10_dr0 = mrcal.compose_r(r1,r0, get_gradients = True)
                    r10_ref = compose_r(r1,r0)

                    if th0 == np.pi and dth == 0. and r0 is r1_simple and r1 is inv_r1_simple_r and iaxis==0:
                        if nps.inner(r10_ref, r10) < 0:
                            # This is about to fail. I'm skipping this test.
                            # It's the only case that fails. And it fails ONLY
                            # on my workstation (recent Debian/sid, Intel(R)
                            # Xeon(R) CPU E5-2687W). It passes on my laptops
                            # (same recent Debian/sid with I think identical
                            # packages, but older CPU: Intel(R) Core(TM)
                            # i7-3520M)
                            print("SKIPPING test case that fails on only some hardware. Everything else passes, so this is likely a subtle roundoff issue")
                            continue

                    confirm_equal( r10,
                                   r10_ref,
                                   worstcase = True,
                                   relative  = True,
                                   eps       = 1e-3,
                                   msg=f'compose_r(r1,r0) near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')
                    dr10_dr0__ref = grad(lambda r0: compose_r(r1,r0),
                                         r0,
                                         forward_differences = True,
                                         switch = wrap_r_unconditional,
                                         step = 1e-7)
                    dr10_dr1__ref = grad(lambda r1: compose_r(r1,r0),
                                         r1,
                                         forward_differences = True,
                                         switch = wrap_r_unconditional,
                                         step = 1e-7)
                    confirm_equal( dr10_dr0,
                                   dr10_dr0__ref,
                                   worstcase = True,
                                   relative  = True,
                                   eps       = 1e-2,
                                   reldiff_eps=1e-3,
                                   msg=f'compose_r(r1,r0 dr10_dr0 near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')
                    confirm_equal( dr10_dr1,
                                   dr10_dr1__ref,
                                   worstcase = True,
                                   relative  = True,
                                   eps       = 1e-2,
                                   reldiff_eps=1e-3,
                                   msg=f'compose_r(r1,r0 dr10_dr1 near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')


            ######### rotate_point_r
            if True:

                # Simple reference implementation. Should move this to
                # test_poseutils_helpers.py. And test_poseutils_lib.py should
                # use it
                def rotate_point_r(r,p):
                    return nps.inner(p, R_from_r(r))

                p = np.array((3., -0.2, -0.9),)

                pt, dpt_dr, dpt_dp = mrcal.rotate_point_r(r,p, get_gradients = True)
                pt_ref             = rotate_point_r(r,p)

                dpt_dr__ref = grad(lambda r: rotate_point_r(r,p),
                                   r)
                dpt_dp__ref = grad(lambda p: rotate_point_r(r,p),
                                   p)
                confirm_equal( pt,
                               pt_ref,
                               msg=f'rotate_point_r(r,p) r near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')
                confirm_equal(dpt_dr,
                              dpt_dr__ref,
                              msg=f'rotate_point_r(r,p) dpt_dr near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')
                confirm_equal(dpt_dp,
                              dpt_dp__ref,
                              msg=f'rotate_point_r(r,p) dpt_dp near a singularity. axis={axis}, th0={th0:.2f}, dth={dth}')

finish()



#!/usr/bin/env python3

r'''Test of the mrcal-graft-models tool

The basic usage of the tool is simple, but it also supports a non-trivial mode
where it applies an implied transformation. I make sure to test that here

'''

import sys
import numpy as np
import numpysane as nps
import os
import subprocess

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils



import tempfile
import atexit
import shutil
workdir = tempfile.mkdtemp()
def cleanup():
    global workdir
    try:
        shutil.rmtree(workdir)
        workdir = None
    except:
        pass
atexit.register(cleanup)



rt_0r = np.array((1.,  2.,    3.,   4., 5.,  6.))
rt_1r = np.array((0.8, -0.01, -0.2, 2., -4., 2.))

imagersize = np.array((2000, 1000), dtype=int)

cxy_center  = (imagersize-1.)/2.
pitch_y     = 100.
cxy_pitched = cxy_center + np.array((0., pitch_y))

# very long lenses. I want rotation to look very much like panning
fxycxy0  = nps.glue( np.array(( 50000., 50000.)), cxy_pitched, axis=-1)
fxycxy1  = nps.glue( np.array(( 50000., 50000.)), cxy_center,  axis=-1)

model0 = mrcal.cameramodel( intrinsics            = ('LENSMODEL_PINHOLE', fxycxy0),
                            imagersize            = imagersize,
                            rt_cam_ref = rt_0r )

model1 = mrcal.cameramodel( intrinsics            = ('LENSMODEL_PINHOLE', fxycxy1),
                            imagersize            = imagersize,
                            rt_cam_ref = rt_1r )

filename0 = f"{workdir}/model0.cameramodel"
filename1 = f"{workdir}/model1.cameramodel"

model0.write(filename0)
model1.write(filename1)

# Basic test. Combine intrinsics and extrinsics without fitting any extra
# transform
out = subprocess.check_output( (f"{testdir}/../mrcal-graft-models",
                                filename0, filename1),
                               encoding = 'ascii',
                               stderr   =  subprocess.DEVNULL)

filename01 = f"{workdir}/model01.cameramodel"
with open(filename01, "w") as f:
    print(out, file=f)

model01 = mrcal.cameramodel(filename01)

testutils.confirm_equal(model01.intrinsics()[1], model0.intrinsics()[1],
                        msg = f"Basic grafted intrinsics match",
                        eps = 1.0e-6)
testutils.confirm_equal(model01.rt_cam_ref(), model1.rt_cam_ref(),
                        msg = f"Basic grafted extrinsics match",
                        eps = 1.0e-6)


# More complicated test. I want to compensate for the different intrinsics with
# modified extrinsics such that the old-intrinsics and new-intrinsics project
# world points to the same place
out = subprocess.check_output( (f"{testdir}/../mrcal-graft-models",
                                '--radius', '-1',
                                filename0, filename1),
                               encoding = 'ascii',
                               stderr   =  subprocess.DEVNULL)

filename01_compensated = f"{workdir}/model01_compensated.cameramodel"
with open(filename01_compensated, "w") as f:
    print(out, file=f)

model01_compensated = mrcal.cameramodel(filename01_compensated)

p1 = np.array((11., 17., 10000.))
pref = mrcal.transform_point_rt( model1.rt_ref_cam(),
                                 p1)

q = mrcal.project( mrcal.transform_point_rt( model1.rt_cam_ref(),
                                             pref ),
                   *model1.intrinsics())
q_compensated = \
    mrcal.project( mrcal.transform_point_rt( model01_compensated.rt_cam_ref(),
                                             pref),
                   *model01_compensated.intrinsics())

testutils.confirm_equal(q_compensated, q,
                        msg = f"Compensated projection ended up in the same place",
                        eps = 0.1)

testutils.finish()



#!/usr/bin/env python3

r'''Checks the C and Python implementations of traverse_sensor_links()
'''

import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils

from mrcal.calibration import _traverse_sensor_links_python


connectivity_matrix = np.array((( 0,1,0,3,0),
                                ( 1,0,2,1,0),
                                ( 0,2,0,0,1),
                                ( 3,1,0,0,0),
                                ( 0,0,1,0,0),),
                               dtype=np.uint16)
Nsensors = connectivity_matrix.shape[0]




node_sequence        = []
node_sequence_python = []

mrcal.traverse_sensor_links(connectivity_matrix = connectivity_matrix,
                                  callback_sensor_link = lambda idx_to,idx_from: node_sequence.append((idx_to,idx_from),))


def cost_edge(camera_idx, from_idx):
    num_shared_frames = connectivity_matrix[camera_idx, from_idx]
    return 65536 - int(num_shared_frames) # int() to case up from uint16, and avoid an overflow
def neighbors(camera_idx):
    for neighbor_idx in range(Nsensors):
        if neighbor_idx == camera_idx                  or \
           connectivity_matrix[neighbor_idx,camera_idx] == 0:
            continue
        yield neighbor_idx


_traverse_sensor_links_python(Nsensors = Nsensors,
                                    callback__neighbors = neighbors,
                                    callback__cost_edge = cost_edge,
                                    callback__sensor_link = lambda idx_to,idx_from: node_sequence_python.append((idx_to,idx_from),))

testutils.confirm_equal( node_sequence,
                         node_sequence_python,
                         worstcase = True,
                         msg=f'The sequence of visited nodes matches in C and Python')

testutils.finish()



r'''
Simple reference implementations of poseutils functions

I compare the mrcal results against these
'''

import sys
import numpy as np
import numpysane as nps
import os
testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal

import scipy.linalg


def R_from_r(r):
    r'''Rotation matrix from a Rodrigues vector

    Simple reference implementation from wikipedia:

    https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
'''

    th_sq = nps.norm2(r)
    Kth = np.array(((    0, -r[2],  r[1]),
                     ( r[2],    0, -r[0]),
                     (-r[1], r[0],     0)))
    if th_sq > 1e-20:
        # normal path
        th = np.sqrt(th_sq)
        s = np.sin(th)
        c = np.cos(th)
        K = Kth  / th
        return np.eye(3) + s*K + (1. - c)*nps.matmult(K,K)

    # small th. Can't divide by it. But I can look at the limit.
    #
    # s*K = Kth * sin(th)/th -> Kth
    #
    # (1-c)*nps.matmult(K,K) = (1-c) Kth^2/th^2 -> Kth^2 s / 2th -> Kth^2/2
    return np.eye(3) + Kth + nps.matmult(Kth,Kth) / 2.


def r_from_R(R):
    r'''Rodrigues vector from a Rotation matrix

    Simple reference implementation from wikipedia:

    https://en.wikipedia.org/wiki/Rotation_matrix#Conversion_from_and_to_axis%E2%80%93angle

    I assume the input is a valid rotation matrix

    '''

    # see the comment in r_from_R_core() for lots of detail

    #       [ R21 - R12 ]
    #   u = [ R02 - R20 ]
    #       [ R10 - R01 ]
    #   u = 2 sin(th) axis
    u = np.array((R[2,1] - R[1,2],
                  R[0,2] - R[2,0],
                  R[1,0] - R[0,1] ))

    costh = (np.trace(R) - 1.)/2.

    if nps.norm2(u) > 1e-12:
        # normal path
        if   costh >  1.0: th = 0
        elif costh < -1.0: th = np.pi
        else:              th = np.arccos(costh)
        return u / nps.mag(u) * th

    # small mag(u). Can't divide by it. But I can look at the limit.
    # th ~ 0 or th ~ 180
    if costh > 0:
        # th ~ 0
        # r = axis th = u / (2 sinth)*th ~ u/2
        return u/2.

    # th ~ 180

    # I need to set rcond because grad() might cause this function to be called
    # with not-quite-rotation matrices
    axis = scipy.linalg.null_space(R - mrcal.identity_R(),
                                   rcond = 1e-6).ravel()
    if axis.size != 3:
        raise Exception("Reference r_from_R implementation did something wrong...")

    # r_from_R_core() has comments. I use this: R - Rt = 2 sin(th) V where V =
    # skew_symmetric(axis)
    #
    # -> sin(th) ~ (R - Rt) / 2V
    #
    # The diagonals of both are 0. Both are anti-symmetric. So I need to look at
    # the 3 off-diagonal elements only.
    #
    # The top-right corner of V is    {-axis[2] axis[1] -axis[0]}
    # The top-right corner of R-Rt is {R01-R10  R02-R20 R12-R21}
    Roffdiag = np.array( (R[0,1]-R[1,0],
                          R[0,2]-R[2,0],
                          R[1,2]-R[2,1]))
    Voffdiag = np.array((-axis[2], axis[1], -axis[0]))
    i = np.abs(Voffdiag) > 0.1
    sinth = np.mean( Roffdiag[i] / Voffdiag[i] ) / 2.
    th = np.arctan2(sinth,costh)
    return axis*th


def Rt_from_rt(rt):
    r'''Simple reference implementation'''
    return nps.glue(R_from_r(rt[:3]), rt[3:], axis=-2)

def rt_from_Rt(Rt):
    r'''Simple reference implementation'''
    return nps.glue(r_from_R(Rt[:3,:]), Rt[3,:], axis=-1)


def invert_rt(rt):
    r'''Simple reference implementation

    b = Ra + t  -> a = R'b - R't
'''
    r = rt[:3]
    t = rt[3:]
    R = R_from_r(r)
    tinv = -nps.matmult(t, R)
    return nps.glue( -r, tinv.ravel(), axis=-1)

def invert_Rt(Rt):
    r'''Simple reference implementation

    b = Ra + t  -> a = R'b - R't
'''
    R = Rt[:3,:]
    tinv = -nps.matmult(Rt[3,:], R)
    return nps.glue( nps.transpose(R), tinv.ravel(), axis=-2)

def invert_R(R):
    r'''Simple reference implementation

'''
    return nps.transpose(R)

def compose_Rt(Rt0, Rt1):
    r'''Simple reference implementation

    b = R0 (R1 x + t1) + t0 =
      = R0 R1 x + R0 t1 + t0
'''
    R0 = Rt0[:3,:]
    t0 = Rt0[ 3,:]
    R1 = Rt1[:3,:]
    t1 = Rt1[ 3,:]
    R2 = nps.matmult(R0,R1)
    t2 = nps.matmult(t1, nps.transpose(R0)) + t0
    return nps.glue( R2, t2.ravel(), axis=-2)

def normalize_r(r):
    r'''If abs(rot) > 180deg -> flip direction, abs(rot) <- 180-abs(rot)'''
    norm2 = nps.norm2(r)
    if norm2 < np.pi*np.pi:
        return r

    mag = np.sqrt(norm2)
    r_unit = r / mag

    mag = mag % (np.pi*2.)
    if mag < np.pi:
        # same direction, but fewer full rotations
        return r_unit*mag

    return -r_unit * (np.pi*2. - mag)

def normalize_rt(rt):
    return nps.glue(normalize_r(rt[:3]),
                    rt[3:],
                    axis=-1)

def compose_r(r0, r1):
    r'''Simple reference implementation'''
    return r_from_R(nps.matmult( R_from_r(r0),
                                 R_from_r(r1)))

def compose_rt(rt0, rt1):
    r'''Simple reference implementation'''
    return rt_from_Rt(compose_Rt( Rt_from_rt(rt0),
                                  Rt_from_rt(rt1)))
